"""Generic class to handle Image-like data in a OME-NGFF file."""

from collections.abc import Mapping, Sequence
from typing import Any, Literal

import dask.array as da
import numpy as np
from pydantic import BaseModel, model_validator
from zarr.core.array import CompressorLike

from ngio.common import (
    Dimensions,
    InterpolationOrder,
    Roi,
)
from ngio.common._pyramid import ChunksLike, ShardsLike
from ngio.images._abstract_image import AbstractImage, abstract_derive
from ngio.io_pipes import (
    SlicingInputType,
    TransformProtocol,
)
from ngio.ome_zarr_meta import (
    ImageMetaHandler,
    NgioImageMeta,
    PixelSize,
)
from ngio.ome_zarr_meta.ngio_specs import (
    Channel,
    ChannelsMeta,
    ChannelVisualisation,
    DefaultSpaceUnit,
    DefaultTimeUnit,
    NgffVersions,
    SpaceUnits,
    TimeUnits,
)
from ngio.utils import (
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)


class ChannelSelectionModel(BaseModel):
    """Model for channel selection.

    This model is used to select a channel by label, wavelength ID, or index.

    Args:
        identifier (str): Unique identifier for the channel.
            This can be a channel label, wavelength ID, or index.
        mode (Literal["label", "wavelength_id", "index"]): Specifies how to
            interpret the identifier. Can be "label", "wavelength_id", or
            "index" (must be an integer).

    """

    mode: Literal["label", "wavelength_id", "index"] = "label"
    identifier: str

    @model_validator(mode="after")
    def check_channel_selection(self):
        if self.mode == "index":
            try:
                int(self.identifier)
            except ValueError as e:
                raise ValueError(
                    "Identifier must be an integer when mode is 'index'"
                ) from e
        return self


ChannelSlicingInputType = (
    None
    | int
    | str
    | ChannelSelectionModel
    | Sequence[str | ChannelSelectionModel | int]
)


def _check_channel_meta(meta: NgioImageMeta, dimension: Dimensions) -> ChannelsMeta:
    """Check the channel metadata."""
    c_dim = dimension.get("c", default=1)

    if meta.channels_meta is None:
        return ChannelsMeta.default_init(labels=c_dim)

    if len(meta.channels_meta.channels) != c_dim:
        raise NgioValueError(
            "The number of channels does not match the image. "
            f"Expected {len(meta.channels_meta.channels)} channels, got {c_dim}."
        )

    return meta.channels_meta


class Image(AbstractImage):
    """A class to handle a single image (or level) in an OME-Zarr image.

    This class is meant to be subclassed by specific image types.
    """

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: ImageMetaHandler | None,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.

        """
        if meta_handler is None:
            meta_handler = ImageMetaHandler(group_handler)
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )

    @property
    def meta_handler(self) -> ImageMetaHandler:
        """Return the metadata handler."""
        assert isinstance(self._meta_handler, ImageMetaHandler)
        return self._meta_handler

    @property
    def meta(self) -> NgioImageMeta:
        """Return the metadata."""
        meta = self.meta_handler.get_meta()
        assert isinstance(meta, NgioImageMeta)
        return meta

    @property
    def channels_meta(self) -> ChannelsMeta:
        """Return the channels metadata."""
        return _check_channel_meta(self.meta, self.dimensions)

    @property
    def channel_labels(self) -> list[str]:
        """Return the channels of the image."""
        return self.channels_meta.channel_labels

    @property
    def wavelength_ids(self) -> list[str | None]:
        """Return the list of wavelength of the image."""
        return self.channels_meta.channel_wavelength_ids

    @property
    def num_channels(self) -> int:
        """Return the number of channels."""
        return len(self.channel_labels)

    def get_channel_idx(
        self, channel_label: str | None = None, wavelength_id: str | None = None
    ) -> int:
        """Get the index of a channel by its label or wavelength ID."""
        return self.channels_meta.get_channel_idx(
            channel_label=channel_label, wavelength_id=wavelength_id
        )

    def get_as_numpy(
        self,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: slice | int | Sequence[int] | None,
    ) -> np.ndarray:
        """Get the image as a numpy array.

        Args:
            channel_selection: Select a specific channel by label.
                If None, all channels are returned.
                Alternatively, you can slice arbitrary channels
                using the slice_kwargs (c=[0, 2]).
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        _slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )
        return self._get_as_numpy(
            axes_order=axes_order, transforms=transforms, **_slicing_kwargs
        )

    def get_roi_as_numpy(
        self,
        roi: Roi,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray:
        """Get the image as a numpy array for a region of interest.

        Args:
            roi: The region of interest to get the array.
            channel_selection: Select a what subset of channels to return.
                If None, all channels are returned.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.

        Returns:
            The array of the region of interest.
        """
        _slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )
        return self._get_roi_as_numpy(
            roi=roi, axes_order=axes_order, transforms=transforms, **_slicing_kwargs
        )

    def get_as_dask(
        self,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> da.Array:
        """Get the image as a dask array.

        Args:
            channel_selection: Select a what subset of channels to return.
                If None, all channels are returned.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.

        Returns:
            The dask array of the region of interest.
        """
        _slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )
        return self._get_as_dask(
            axes_order=axes_order, transforms=transforms, **_slicing_kwargs
        )

    def get_roi_as_dask(
        self,
        roi: Roi,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> da.Array:
        """Get the image as a dask array for a region of interest.

        Args:
            roi: The region of interest to get the array.
            channel_selection: Select a what subset of channels to return.
                If None, all channels are returned.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to get the array.

        Returns:
            The dask array of the region of interest.
        """
        _slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )
        return self._get_roi_as_dask(
            roi=roi, axes_order=axes_order, transforms=transforms, **_slicing_kwargs
        )

    def get_array(
        self,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray | da.Array:
        """Get the image as a zarr array.

        Args:
            channel_selection: Select a what subset of channels to return.
                If None, all channels are returned.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            mode: The object type to return.
                Can be "dask", "numpy".
            **slicing_kwargs: The slices to get the array.

        Returns:
            The zarr array of the region of interest.
        """
        _slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )
        return self._get_array(
            axes_order=axes_order, mode=mode, transforms=transforms, **_slicing_kwargs
        )

    def get_roi(
        self,
        roi: Roi,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        mode: Literal["numpy", "dask"] = "numpy",
        **slicing_kwargs: SlicingInputType,
    ) -> np.ndarray | da.Array:
        """Get the image as a zarr array for a region of interest.

        Args:
            roi: The region of interest to get the array.
            channel_selection: Select a what subset of channels to return.
                If None, all channels are returned.
            axes_order: The order of the axes to return the array.
            transforms: The transforms to apply to the array.
            mode: The object type to return.
                Can be "dask", "numpy".
            **slicing_kwargs: The slices to get the array.

        Returns:
            The zarr array of the region of interest.
        """
        _slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )
        return self._get_roi(
            roi=roi,
            axes_order=axes_order,
            mode=mode,
            transforms=transforms,
            **_slicing_kwargs,
        )

    def set_array(
        self,
        patch: np.ndarray | da.Array,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> None:
        """Set the image array.

        Args:
            patch: The array to set.
            channel_selection: Select a what subset of channels to return.
                If None, all channels are set.
            axes_order: The order of the axes to set the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to set the array.
        """
        _slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )
        self._set_array(
            patch=patch, axes_order=axes_order, transforms=transforms, **_slicing_kwargs
        )

    def set_roi(
        self,
        roi: Roi,
        patch: np.ndarray | da.Array,
        channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        transforms: Sequence[TransformProtocol] | None = None,
        **slicing_kwargs: SlicingInputType,
    ) -> None:
        """Set the image array for a region of interest.

        Args:
            roi: The region of interest to set the array.
            patch: The array to set.
            channel_selection: Select a what subset of channels to return.
            axes_order: The order of the axes to set the array.
            transforms: The transforms to apply to the array.
            **slicing_kwargs: The slices to set the array.
        """
        _slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self, channel_selection=channel_selection, slicing_dict=slicing_kwargs
        )
        self._set_roi(
            roi=roi,
            patch=patch,
            axes_order=axes_order,
            transforms=transforms,
            **_slicing_kwargs,
        )

    def consolidate(
        self,
        order: InterpolationOrder = "linear",
        mode: Literal["dask", "numpy", "coarsen"] = "dask",
    ) -> None:
        """Consolidate the label on disk."""
        self._consolidate(order=order, mode=mode)


class ImagesContainer:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(self, group_handler: ZarrGroupHandler) -> None:
        """Initialize the LabelGroupHandler."""
        self._group_handler = group_handler
        self._meta_handler = ImageMetaHandler(group_handler)

    @property
    def meta(self) -> NgioImageMeta:
        """Return the metadata."""
        return self._meta_handler.get_meta()

    @property
    def levels(self) -> int:
        """Return the number of levels in the image."""
        return self._meta_handler.get_meta().levels

    @property
    def levels_paths(self) -> list[str]:
        """Return the paths of the levels in the image."""
        return self._meta_handler.get_meta().paths

    @property
    def num_channels(self) -> int:
        """Return the number of channels."""
        image = self.get()
        return image.num_channels

    @property
    def channel_labels(self) -> list[str]:
        """Return the channels of the image."""
        image = self.get()
        return image.channel_labels

    @property
    def wavelength_ids(self) -> list[str | None]:
        """Return the wavelength of the image."""
        image = self.get()
        return image.wavelength_ids

    def get_channel_idx(
        self, channel_label: str | None = None, wavelength_id: str | None = None
    ) -> int:
        """Get the index of a channel by label or wavelength ID.

        Args:
            channel_label (str | None): The label of the channel.
                If None a wavelength ID must be provided.
            wavelength_id (str | None): The wavelength ID of the channel.
                If None a channel label must be provided.

        Returns:
            int: The index of the channel.

        """
        image = self.get()
        return image.get_channel_idx(
            channel_label=channel_label, wavelength_id=wavelength_id
        )

    def _set_channel_meta(
        self,
        channels_meta: ChannelsMeta,
    ) -> None:
        """Set the channels metadata."""
        meta = self.meta
        meta.set_channels_meta(channels_meta)
        self._meta_handler.update_meta(meta)

    def set_channel_meta(
        self,
        labels: Sequence[str | None] | int | None = None,
        wavelength_id: Sequence[str | None] | None = None,
        start: Sequence[float | None] | None = None,
        end: Sequence[float | None] | None = None,
        percentiles: tuple[float, float] | None = None,
        colors: Sequence[str | None] | None = None,
        active: Sequence[bool | None] | None = None,
        **omero_kwargs: dict,
    ) -> None:
        """Create a ChannelsMeta object with the default unit.

        Args:
            labels(Sequence[str | None] | int): The list of channels names
                in the image. If an integer is provided, the channels will
                be named "channel_i".
            wavelength_id(Sequence[str | None]): The wavelength ID of the channel.
                If None, the wavelength ID will be the same as the channel name.
            start(Sequence[float | None]): The start value for each channel.
                If None, the start value will be computed from the image.
            end(Sequence[float | None]): The end value for each channel.
                If None, the end value will be computed from the image.
            percentiles(tuple[float, float] | None): The start and end
                percentiles for each channel. If None, the percentiles will
                not be computed.
            colors(Sequence[str | None]): The list of colors for the
                channels. If None, the colors will be random.
            active (Sequence[bool | None]): Whether the channel should
                be shown by default.
            omero_kwargs(dict): Extra fields to store in the omero attributes.
        """
        low_res_dataset = self.meta.get_lowest_resolution_dataset()
        ref_image = self.get(path=low_res_dataset.path)

        if start is not None and end is None:
            raise NgioValueError("If start is provided, end must be provided as well.")
        if end is not None and start is None:
            raise NgioValueError("If end is provided, start must be provided as well.")

        if start is not None and percentiles is not None:
            raise NgioValueError(
                "If start and end are provided, percentiles must be None."
            )

        if percentiles is not None:
            start, end = compute_image_percentile(
                ref_image,
                start_percentile=percentiles[0],
                end_percentile=percentiles[1],
            )
        elif start is not None and end is not None:
            if len(start) != len(end):
                raise NgioValueError(
                    "The start and end lists must have the same length."
                )
            if len(start) != self.num_channels:
                raise NgioValueError(
                    "The start and end lists must have the same length as "
                    "the number of channels."
                )

            start = list(start)
            end = list(end)

        else:
            start, end = None, None

        if labels is None:
            labels = ref_image.num_channels

        channel_meta = ChannelsMeta.default_init(
            labels=labels,
            wavelength_id=wavelength_id,
            colors=colors,
            start=start,
            end=end,
            active=active,
            data_type=ref_image.dtype,
            **omero_kwargs,
        )
        self._set_channel_meta(channel_meta)

    def set_channel_percentiles(
        self,
        start_percentile: float = 0.1,
        end_percentile: float = 99.9,
    ) -> None:
        """Update the percentiles of the channels."""
        if self.meta._channels_meta is None:
            raise NgioValueError("The channels meta is not initialized.")

        low_res_dataset = self.meta.get_lowest_resolution_dataset()
        ref_image = self.get(path=low_res_dataset.path)
        starts, ends = compute_image_percentile(
            ref_image, start_percentile=start_percentile, end_percentile=end_percentile
        )

        channels = []
        for c, channel in enumerate(self.meta._channels_meta.channels):
            new_v = ChannelVisualisation(
                start=starts[c],
                end=ends[c],
                **channel.channel_visualisation.model_dump(exclude={"start", "end"}),
            )
            new_c = Channel(
                channel_visualisation=new_v,
                **channel.model_dump(exclude={"channel_visualisation"}),
            )
            channels.append(new_c)

        new_meta = ChannelsMeta(channels=channels)

        meta = self.meta
        meta.set_channels_meta(new_meta)
        self._meta_handler.update_meta(meta)

    def set_axes_unit(
        self,
        space_unit: SpaceUnits = DefaultSpaceUnit,
        time_unit: TimeUnits = DefaultTimeUnit,
    ) -> None:
        """Set the axes unit of the image.

        Args:
            space_unit (SpaceUnits): The space unit of the image.
            time_unit (TimeUnits): The time unit of the image.
        """
        meta = self.meta
        meta = meta.to_units(space_unit=space_unit, time_unit=time_unit)
        self._meta_handler.update_meta(meta)

    def derive(
        self,
        store: StoreOrGroup,
        ref_path: str | None = None,
        # Metadata parameters
        shape: Sequence[int] | None = None,
        pixelsize: float | tuple[float, float] | None = None,
        z_spacing: float | None = None,
        time_spacing: float | None = None,
        name: str | None = None,
        channels_meta: Sequence[str | Channel] | None = None,
        ngff_version: NgffVersions | None = None,
        # Zarr Array parameters
        chunks: ChunksLike = "auto",
        shards: ShardsLike | None = None,
        dtype: str = "uint16",
        dimension_separator: Literal[".", "/"] = "/",
        compressors: CompressorLike = "auto",
        extra_array_kwargs: Mapping[str, Any] | None = None,
        overwrite: bool = False,
        # Deprecated arguments
        labels: Sequence[str] | None = None,
        pixel_size: PixelSize | None = None,
    ) -> "ImagesContainer":
        """Create an empty OME-Zarr image from an existing image.

        If a kwarg is not provided, the value from the reference image will be used.

        Args:
            image_container (ImagesContainer): The image container to derive the new
                image.
            store (StoreOrGroup): The Zarr store or group to create the image in.
            ref_path (str | None): The path to the reference image in the image
                container.
            shape (Sequence[int] | None): The shape of the new image.
            pixelsize (float | tuple[float, float] | None): The pixel size of the new
                image.
            z_spacing (float | None): The z spacing of the new image.
            time_spacing (float | None): The time spacing of the new image.
            name (str | None): The name of the new image.
            channels_meta (Sequence[str | Channel] | None): The channels metadata
                of the new image.
            ngff_version (NgffVersions | None): The NGFF version to use.
            chunks (Sequence[int] | None): The chunk shape of the new image.
            shards (ShardsLike | None): The shard shape of the new image.
            dtype (str | None): The data type of the new image.
            dimension_separator (DIMENSION_SEPARATOR | None): The separator to use for
                dimensions.
            compressors (CompressorLike | None): The compressors to use.
            extra_array_kwargs (Mapping[str, Any] | None): Extra arguments to pass to
                the zarr array creation.
            overwrite (bool): Whether to overwrite an existing image.
            labels (Sequence[str] | None): The labels of the new image.
                This argument is deprecated please use channels_meta instead.
            pixel_size (PixelSize | None): The pixel size of the new image.
                This argument is deprecated please use pixelsize, z_spacing,
                and time_spacing instead.

        Returns:
            ImagesContainer: The new derived image.

        """
        return derive_image_container(
            image_container=self,
            store=store,
            ref_path=ref_path,
            shape=shape,
            pixelsize=pixelsize,
            z_spacing=z_spacing,
            time_spacing=time_spacing,
            name=name,
            channels_meta=channels_meta,
            ngff_version=ngff_version,
            chunks=chunks,
            shards=shards,
            dtype=dtype,
            dimension_separator=dimension_separator,
            compressors=compressors,
            extra_array_kwargs=extra_array_kwargs,
            overwrite=overwrite,
            labels=labels,
            pixel_size=pixel_size,
        )

    def get(
        self,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        strict: bool = False,
    ) -> Image:
        """Get an image at a specific level.

        Args:
            path (str | None): The path to the image in the ome_zarr file.
            pixel_size (PixelSize | None): The pixel size of the image.
            strict (bool): Only used if the pixel size is provided. If True, the
                pixel size must match the image pixel size exactly. If False, the
                closest pixel size level will be returned.

        """
        dataset = self._meta_handler.get_meta().get_dataset(
            path=path, pixel_size=pixel_size, strict=strict
        )
        return Image(
            group_handler=self._group_handler,
            path=dataset.path,
            meta_handler=self._meta_handler,
        )


def compute_image_percentile(
    image: Image,
    start_percentile: float = 0.1,
    end_percentile: float = 99.9,
) -> tuple[list[float], list[float]]:
    """Compute the start and end percentiles for each channel of an image.

    Args:
        image: The image to compute the percentiles for.
        start_percentile: The start percentile to compute.
        end_percentile: The end percentile to compute.

    Returns:
        A tuple containing the start and end percentiles for each channel.
    """
    starts, ends = [], []
    for c in range(image.num_channels):
        if image.num_channels == 1:
            data = image.get_as_dask()
        else:
            data = image.get_as_dask(c=c)

        data = da.ravel(data)
        # remove all the zeros
        mask = data > 1e-16
        data = data[mask]
        _data = data.compute()
        if _data.size == 0:
            starts.append(0.0)
            ends.append(0.0)
            continue

        # compute the percentiles
        _s_perc, _e_perc = da.percentile(
            data, [start_percentile, end_percentile], method="nearest"
        ).compute()  # type: ignore (return type is a tuple of floats)

        starts.append(float(_s_perc))
        ends.append(float(_e_perc))
    return starts, ends


def derive_image_container(
    *,
    image_container: ImagesContainer,
    store: StoreOrGroup,
    ref_path: str | None = None,
    # Metadata parameters
    shape: Sequence[int] | None = None,
    pixelsize: float | tuple[float, float] | None = None,
    z_spacing: float | None = None,
    time_spacing: float | None = None,
    name: str | None = None,
    channels_policy: Literal["same", "squeeze", "singleton"] | int = "same",
    channels_meta: Sequence[str | Channel] | None = None,
    ngff_version: NgffVersions | None = None,
    # Zarr Array parameters
    chunks: ChunksLike = "auto",
    shards: ShardsLike | None = None,
    dtype: str | None = None,
    dimension_separator: Literal[".", "/"] | None = None,
    compressors: CompressorLike | None = None,
    extra_array_kwargs: Mapping[str, Any] | None = None,
    overwrite: bool = False,
    # Deprecated arguments
    labels: Sequence[str] | None = None,
    pixel_size: PixelSize | None = None,
) -> ImagesContainer:
    """Derive a new OME-Zarr image container from an existing image.

    If a kwarg is not provided, the value from the reference image will be used.

    Args:
        image_container (ImagesContainer): The image container to derive the new image
            from.
        store (StoreOrGroup): The Zarr store or group to create the image in.
        ref_path (str | None): The path to the reference image in the image container.
        shape (Sequence[int] | None): The shape of the new image.
        pixelsize (float | tuple[float, float] | None): The pixel size of the new image.
        z_spacing (float | None): The z spacing of the new image.
        time_spacing (float | None): The time spacing of the new image.
        name (str | None): The name of the new image.
        channels_policy (Literal["squeeze", "same", "singleton"] | int): Possible
            policies:
            - If "squeeze", the channels axis will be removed (no matter its size).
            - If "same", the channels axis will be kept as is (if it exists).
            - If "singleton", the channels axis will be set to size 1.
            - If an integer is provided, the channels axis will be changed to have
                that size.
        channels_meta (Sequence[str | Channel] | None): The channels metadata
            of the new image.
        ngff_version (NgffVersions | None): The NGFF version to use.
        chunks (ChunksLike): The chunk shape of the new image. Defaults to "auto".
        shards (ShardsLike | None): The shard shape of the new image.
        dtype (str | None): The data type of the new image.
        dimension_separator (Literal[".", "/"] | None): The separator to use for
            dimensions.
        compressors (CompressorLike | None): The compressors to use.
        extra_array_kwargs (Mapping[str, Any] | None): Extra arguments to pass to
            the zarr array creation.
        overwrite (bool): Whether to overwrite an existing image. Defaults to False.
        labels (Sequence[str] | None): Deprecated. This argument is deprecated,
            please use channels_meta instead.
        pixel_size (PixelSize | None): Deprecated. The pixel size of the new image.
            This argument is deprecated, please use pixelsize, z_spacing,
            and time_spacing instead.

    Returns:
        ImagesContainer: The new derived image container.

    """
    ref_image = image_container.get(path=ref_path)
    group_handler = abstract_derive(
        ref_image=ref_image,
        meta_type=NgioImageMeta,
        store=store,
        shape=shape,
        pixelsize=pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        name=name,
        channels_meta=channels_meta,
        channels_policy=channels_policy,
        ngff_version=ngff_version,
        chunks=chunks,
        shards=shards,
        dtype=dtype,
        dimension_separator=dimension_separator,
        compressors=compressors,
        extra_array_kwargs=extra_array_kwargs,
        overwrite=overwrite,
        labels=labels,
        pixel_size=pixel_size,
    )
    return ImagesContainer(group_handler=group_handler)


def _parse_str_or_model(
    image: Image, channel_selection: int | str | ChannelSelectionModel
) -> int:
    """Parse a string or ChannelSelectionModel to an integer channel index."""
    if isinstance(channel_selection, int):
        if channel_selection < 0:
            raise NgioValueError("Channel index must be a non-negative integer.")
        if channel_selection >= image.num_channels:
            raise NgioValueError(
                "Channel index must be less than the number "
                f"of channels ({image.num_channels})."
            )
        return channel_selection
    elif isinstance(channel_selection, str):
        return image.get_channel_idx(channel_label=channel_selection)
    elif isinstance(channel_selection, ChannelSelectionModel):
        if channel_selection.mode == "label":
            return image.get_channel_idx(
                channel_label=str(channel_selection.identifier)
            )
        elif channel_selection.mode == "wavelength_id":
            return image.get_channel_idx(
                wavelength_id=str(channel_selection.identifier)
            )
        elif channel_selection.mode == "index":
            return int(channel_selection.identifier)
    raise NgioValueError(
        "Invalid channel selection type. "
        f"{channel_selection} is of type {type(channel_selection)} ",
        "supported types are str, ChannelSelectionModel, and int.",
    )


def _parse_channel_selection(
    image: Image, channel_selection: ChannelSlicingInputType
) -> dict[str, SlicingInputType]:
    """Parse the channel selection input into a list of channel indices."""
    if channel_selection is None:
        return {}
    if isinstance(channel_selection, int | str | ChannelSelectionModel):
        channel_index = _parse_str_or_model(image, channel_selection)
        return {"c": channel_index}
    elif isinstance(channel_selection, Sequence):
        _sequence = [_parse_str_or_model(image, cs) for cs in channel_selection]
        return {"c": _sequence}
    raise NgioValueError(
        f"Invalid channel selection type {type(channel_selection)}. "
        "Supported types are int, str, ChannelSelectionModel, and Sequence."
    )


def add_channel_selection_to_slicing_dict(
    image: Image,
    channel_selection: ChannelSlicingInputType,
    slicing_dict: dict[str, SlicingInputType],
) -> dict[str, SlicingInputType]:
    """Add channel selection information to the slicing dictionary."""
    channel_info = _parse_channel_selection(image, channel_selection)
    if "c" in slicing_dict and channel_info:
        raise NgioValueError(
            "Both channel_selection and 'c' in slicing_kwargs are provided. "
            "Which channel selection should be used is ambiguous. "
            "Please provide only one."
        )
    slicing_dict = slicing_dict | channel_info
    return slicing_dict



from typing import Generic, TypeVar

T = TypeVar("T")


class NgioCache(Generic[T]):
    """A simple cache for NGIO objects."""

    def __init__(self, use_cache: bool = True):
        self._cache: dict[str, T] = {}
        self._use_cache = use_cache

    def _cache_sanity_check(self) -> None:
        if len(self._cache) > 0:
            raise RuntimeError(
                "Cache is disabled, but cache contains items. "
                "This indicates a logic error."
            )

    @property
    def use_cache(self) -> bool:
        return self._use_cache

    @property
    def cache(self) -> dict[str, T]:
        return self._cache

    @property
    def is_empty(self) -> bool:
        return len(self._cache) == 0

    def get(self, key: str, default: T | None = None) -> T | None:
        if not self._use_cache:
            self._cache_sanity_check()
            return default
        return self._cache.get(key, default)

    def set(self, key: str, value: T, overwrite: bool = True) -> None:
        if not self._use_cache:
            self._cache_sanity_check()
            return
        self._cache[key] = value

    def clear(self) -> None:
        if not self._use_cache:
            self._cache_sanity_check()
            return
        self._cache.clear()



"""Utilities for reading and writing OME-Zarr metadata."""

from ngio.ome_zarr_meta._meta_handlers import (
    ImageMetaHandler,
    LabelMetaHandler,
    LabelsGroupMetaHandler,
    PlateMetaHandler,
    WellMetaHandler,
    update_ngio_image_meta,
    update_ngio_label_meta,
    update_ngio_labels_group_meta,
    update_ngio_meta,
    update_ngio_plate_meta,
    update_ngio_well_meta,
)
from ngio.ome_zarr_meta.ngio_specs import (
    AxesHandler,
    Dataset,
    ImageInWellPath,
    NgffVersions,
    NgioImageMeta,
    NgioLabelMeta,
    NgioLabelsGroupMeta,
    NgioPlateMeta,
    NgioWellMeta,
    PixelSize,
    build_canonical_axes_handler,
    path_in_well_validation,
)

__all__ = [
    "AxesHandler",
    "Dataset",
    "ImageInWellPath",
    "ImageMetaHandler",
    "LabelMetaHandler",
    "LabelsGroupMetaHandler",
    "NgffVersions",
    "NgffVersions",
    "NgioImageMeta",
    "NgioLabelMeta",
    "NgioLabelsGroupMeta",
    "NgioPlateMeta",
    "NgioWellMeta",
    "PixelSize",
    "PlateMetaHandler",
    "PlateMetaHandler",
    "WellMetaHandler",
    "build_canonical_axes_handler",
    "path_in_well_validation",
    "update_ngio_image_meta",
    "update_ngio_label_meta",
    "update_ngio_labels_group_meta",
    "update_ngio_meta",
    "update_ngio_plate_meta",
    "update_ngio_well_meta",
]



# conftest.py
import os
import socket
import subprocess
import sys
import time
from pathlib import Path

import boto3
import pytest


def _running_on_github_ci() -> bool:
    return os.getenv("GITHUB_ACTIONS") == "true" or os.getenv("CI") == "true"


if sys.platform == "darwin" and _running_on_github_ci():
    # The store tests require local servers which seem to have issues on macOS CI.
    # Skip the whole module in this case.
    pytestmark = pytest.skip(
        reason="Integration tests (local servers) are skipped on macOS CI",
        allow_module_level=True,
    )


def _wait_for_port(proc, host, port, timeout=20):
    """Wait until a TCP port starts accepting connections, or the process dies."""
    start = time.time()
    while time.time() - start < timeout:
        # bail out early if the process crashed
        if proc.poll() is not None:
            raise RuntimeError("moto server process exited early")

        try:
            with socket.create_connection((host, port), timeout=0.5):
                return
        except OSError:
            time.sleep(0.2)

    raise RuntimeError(f"Port {port} did not open in time")


@pytest.fixture(scope="session")
def moto_s3_server(tmp_path_factory):
    host = "127.0.0.1"
    port = 5005

    log_dir = tmp_path_factory.mktemp("moto_logs")
    log_file = log_dir / "server.log"

    env = os.environ.copy()
    env["PYTHONUNBUFFERED"] = "1"
    # only start S3 service f
    env["MOTO_SERVICE"] = "s3"

    # NOTE: no "s3" argument here anymore
    cmd = [
        sys.executable,
        "-m",
        "moto.server",
        "-p",
        str(port),
    ]

    with log_file.open("wb") as lf:
        proc = subprocess.Popen(
            cmd,
            stdout=lf,
            stderr=subprocess.STDOUT,
            env=env,
        )

    try:
        _wait_for_port(proc, host, port, timeout=20)
    except Exception as e:
        proc.terminate()
        try:
            proc.wait(timeout=5)
        except subprocess.TimeoutExpired:
            proc.kill()
        log_text = log_file.read_text(errors="replace")
        raise RuntimeError(
            f"Failed to start moto server: {e}\n--- moto log ---\n{log_text}"
        ) from e

    s3_endpoint_url = f"http://{host}:{port}"
    bucket_name = "s3-ci-test-bucket"
    s3 = boto3.client(
        "s3",
        region_name="us-east-1",
        aws_access_key_id="test",
        aws_secret_access_key="test",
        endpoint_url=s3_endpoint_url,
    )
    s3.create_bucket(Bucket=bucket_name)
    yield {"endpoint_url": s3_endpoint_url, "bucket_name": bucket_name}

    proc.terminate()
    try:
        proc.wait(timeout=5)
    except subprocess.TimeoutExpired:
        proc.kill()


def _find_free_port(host="127.0.0.1"):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind((host, 0))
        return s.getsockname()[1]


@pytest.fixture(scope="session")
def http_static_server(tmp_path_factory):
    """
    Serve a temporary directory via `python -m http.server`.

    From the test code's perspective this is read-only: you write files
    directly into `root` on disk, and then access them via HTTP.
    """
    host = "127.0.0.1"
    port = _find_free_port(host)

    root = tmp_path_factory.mktemp("http_static_root")

    cmd = [
        sys.executable,
        "-m",
        "http.server",
        str(port),
        "--bind",
        host,
    ]

    env = os.environ.copy()
    env["PYTHONUNBUFFERED"] = "1"

    proc = subprocess.Popen(
        cmd,
        cwd=str(root),  # serve this directory
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        env=env,
    )

    try:
        _wait_for_port(proc, host, port, timeout=10)
    except Exception as e:
        proc.terminate()
        try:
            proc.wait(timeout=5)
        except subprocess.TimeoutExpired:
            proc.kill()
        raise RuntimeError(f"Failed to start http server: {e}") from e

    yield {"url": f"http://{host}:{port}", "root": Path(root)}

    proc.terminate()
    try:
        proc.wait(timeout=5)
    except subprocess.TimeoutExpired:
        proc.kill()



"""Ngio Tables backend implementations."""

from ngio.tables.backends._abstract_backend import AbstractTableBackend, BackendMeta
from ngio.tables.backends._anndata import AnnDataBackend, AnnDataBackendV1
from ngio.tables.backends._csv import CsvTableBackend
from ngio.tables.backends._json import JsonTableBackend
from ngio.tables.backends._parquet import ParquetTableBackend
from ngio.tables.backends._table_backends import (
    DefaultTableBackend,
    ImplementedTableBackends,
    TableBackend,
    TableBackendProtocol,
)
from ngio.tables.backends._utils import (
    TabularData,
    convert_anndata_to_pandas,
    convert_anndata_to_polars,
    convert_pandas_to_anndata,
    convert_pandas_to_polars,
    convert_polars_to_anndata,
    convert_polars_to_pandas,
    convert_to_anndata,
    convert_to_pandas,
    convert_to_polars,
    normalize_anndata,
    normalize_pandas_df,
    normalize_polars_lf,
    normalize_table,
)

__all__ = [
    "AbstractTableBackend",
    "AnnDataBackend",
    "AnnDataBackendV1",
    "BackendMeta",
    "CsvTableBackend",
    "DefaultTableBackend",
    "ImplementedTableBackends",
    "JsonTableBackend",
    "ParquetTableBackend",
    "TableBackend",
    "TableBackendProtocol",
    "TabularData",
    "convert_anndata_to_pandas",
    "convert_anndata_to_polars",
    "convert_pandas_to_anndata",
    "convert_pandas_to_polars",
    "convert_polars_to_anndata",
    "convert_polars_to_pandas",
    "convert_to_anndata",
    "convert_to_pandas",
    "convert_to_polars",
    "normalize_anndata",
    "normalize_pandas_df",
    "normalize_polars_lf",
    "normalize_table",
]



from collections.abc import Sequence

import dask.array as da
import numpy as np

from ngio.common import Roi
from ngio.experimental.iterators._abstract_iterator import AbstractIteratorBuilder
from ngio.images import Image
from ngio.images._image import (
    ChannelSlicingInputType,
    add_channel_selection_to_slicing_dict,
)
from ngio.io_pipes import (
    DaskRoiGetter,
    DaskRoiSetter,
    NumpyRoiGetter,
    NumpyRoiSetter,
    TransformProtocol,
)
from ngio.io_pipes._io_pipes_types import DataGetterProtocol, DataSetterProtocol


class ImageProcessingIterator(AbstractIteratorBuilder[np.ndarray, da.Array]):
    """Base class for iterators over ROIs."""

    def __init__(
        self,
        input_image: Image,
        output_image: Image,
        input_channel_selection: ChannelSlicingInputType = None,
        output_channel_selection: ChannelSlicingInputType = None,
        axes_order: Sequence[str] | None = None,
        input_transforms: Sequence[TransformProtocol] | None = None,
        output_transforms: Sequence[TransformProtocol] | None = None,
    ) -> None:
        """Initialize the iterator with a ROI table and input/output images.

        Args:
            input_image (Image): The input image to be used as input for the
                segmentation.
            output_image (Image): The image where the ROIs will be written.
            input_channel_selection (ChannelSlicingInputType): Optional
                selection of channels to use for the input image.
            output_channel_selection (ChannelSlicingInputType): Optional
                selection of channels to use for the output image.
            axes_order (Sequence[str] | None): Optional axes order for the
                segmentation.
            input_transforms (Sequence[TransformProtocol] | None): Optional
                transforms to apply to the input image.
            output_transforms (Sequence[TransformProtocol] | None): Optional
                transforms to apply to the output label.
        """
        self._input = input_image
        self._output = output_image
        self._ref_image = input_image
        self._rois = input_image.build_image_roi_table(name=None).rois()

        # Set iteration parameters
        self._input_slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self._input,
            channel_selection=input_channel_selection,
            slicing_dict={},
        )
        self._output_slicing_kwargs = add_channel_selection_to_slicing_dict(
            image=self._output,
            channel_selection=output_channel_selection,
            slicing_dict={},
        )
        self._input_channel_selection = input_channel_selection
        self._output_channel_selection = output_channel_selection
        self._axes_order = axes_order
        self._input_transforms = input_transforms
        self._output_transforms = output_transforms

        self._input.require_dimensions_match(self._output, allow_singleton=True)

    def get_init_kwargs(self) -> dict:
        """Return the initialization arguments for the iterator."""
        return {
            "input_image": self._input,
            "output_image": self._output,
            "input_channel_selection": self._input_channel_selection,
            "output_channel_selection": self._output_channel_selection,
            "axes_order": self._axes_order,
            "input_transforms": self._input_transforms,
            "output_transforms": self._output_transforms,
        }

    def build_numpy_getter(self, roi: Roi) -> DataGetterProtocol[np.ndarray]:
        return NumpyRoiGetter(
            zarr_array=self._input.zarr_array,
            dimensions=self._input.dimensions,
            roi=roi,
            axes_order=self._axes_order,
            transforms=self._input_transforms,
            slicing_dict=self._input_slicing_kwargs,
        )

    def build_numpy_setter(self, roi: Roi) -> DataSetterProtocol[np.ndarray]:
        return NumpyRoiSetter(
            zarr_array=self._output.zarr_array,
            dimensions=self._output.dimensions,
            roi=roi,
            axes_order=self._axes_order,
            transforms=self._output_transforms,
            slicing_dict=self._output_slicing_kwargs,
        )

    def build_dask_getter(self, roi: Roi) -> DataGetterProtocol[da.Array]:
        return DaskRoiGetter(
            zarr_array=self._input.zarr_array,
            dimensions=self._input.dimensions,
            roi=roi,
            axes_order=self._axes_order,
            transforms=self._input_transforms,
            slicing_dict=self._input_slicing_kwargs,
        )

    def build_dask_setter(self, roi: Roi) -> DataSetterProtocol[da.Array]:
        return DaskRoiSetter(
            zarr_array=self._output.zarr_array,
            dimensions=self._output.dimensions,
            roi=roi,
            axes_order=self._axes_order,
            transforms=self._output_transforms,
            slicing_dict=self._output_slicing_kwargs,
        )

    def post_consolidate(self):
        self._output.consolidate()



"""A module for handling label images in OME-NGFF files."""

from collections.abc import Mapping, Sequence
from typing import Any, Literal

from zarr.core.array import CompressorLike

from ngio.common import compute_masking_roi
from ngio.common._pyramid import ChunksLike, ShardsLike
from ngio.images._abstract_image import AbstractImage, abstract_derive
from ngio.images._image import Image
from ngio.ome_zarr_meta import (
    LabelMetaHandler,
    LabelsGroupMetaHandler,
    NgioLabelMeta,
    NgioLabelsGroupMeta,
    PixelSize,
    update_ngio_labels_group_meta,
)
from ngio.ome_zarr_meta.ngio_specs import (
    DefaultSpaceUnit,
    DefaultTimeUnit,
    NgffVersions,
    SpaceUnits,
    TimeUnits,
)
from ngio.tables import MaskingRoiTable
from ngio.utils import (
    NgioValidationError,
    NgioValueError,
    StoreOrGroup,
    ZarrGroupHandler,
)


class Label(AbstractImage):
    """Placeholder class for a label."""

    get_as_numpy = AbstractImage._get_as_numpy
    get_as_dask = AbstractImage._get_as_dask
    get_array = AbstractImage._get_array
    get_roi_as_numpy = AbstractImage._get_roi_as_numpy
    get_roi_as_dask = AbstractImage._get_roi_as_dask
    get_roi = AbstractImage._get_roi
    set_array = AbstractImage._set_array
    set_roi = AbstractImage._set_roi

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        path: str,
        meta_handler: LabelMetaHandler | None,
    ) -> None:
        """Initialize the Image at a single level.

        Args:
            group_handler: The Zarr group handler.
            path: The path to the image in the ome_zarr file.
            meta_handler: The image metadata handler.

        """
        if meta_handler is None:
            meta_handler = LabelMetaHandler(group_handler)
        super().__init__(
            group_handler=group_handler, path=path, meta_handler=meta_handler
        )

    def __repr__(self) -> str:
        """Return the string representation of the label."""
        return f"Label(path={self.path}, {self.dimensions})"

    @property
    def meta_handler(self) -> LabelMetaHandler:
        """Return the metadata handler."""
        assert isinstance(self._meta_handler, LabelMetaHandler)
        return self._meta_handler

    @property
    def meta(self) -> NgioLabelMeta:
        """Return the metadata."""
        meta = self.meta_handler.get_meta()
        assert isinstance(meta, NgioLabelMeta)
        return meta

    def set_axes_unit(
        self,
        space_unit: SpaceUnits = DefaultSpaceUnit,
        time_unit: TimeUnits = DefaultTimeUnit,
    ) -> None:
        """Set the axes unit of the image.

        Args:
            space_unit (SpaceUnits): The space unit of the image.
            time_unit (TimeUnits): The time unit of the image.
        """
        meta = self.meta
        meta = meta.to_units(space_unit=space_unit, time_unit=time_unit)
        self.meta_handler.update_meta(meta)

    def build_masking_roi_table(self) -> MaskingRoiTable:
        """Compute the masking ROI table."""
        return build_masking_roi_table(self)

    def consolidate(
        self,
        mode: Literal["dask", "numpy", "coarsen"] = "dask",
    ) -> None:
        """Consolidate the label on disk."""
        self._consolidate(
            order="nearest",
            mode=mode,
        )


class LabelsContainer:
    """A class to handle the /labels group in an OME-NGFF file."""

    def __init__(
        self,
        group_handler: ZarrGroupHandler,
        ngff_version: NgffVersions | None = None,
    ) -> None:
        """Initialize the LabelGroupHandler."""
        self._group_handler = group_handler
        # If the group is empty, initialize the metadata
        try:
            self._meta_handler = LabelsGroupMetaHandler(group_handler)
        except NgioValidationError:
            if ngff_version is None:
                raise NgioValueError(
                    "The /labels group is missing metadata. "
                    "Please provide the ngff_version to initialize it."
                ) from None
            meta = NgioLabelsGroupMeta(labels=[], version=ngff_version)
            update_ngio_labels_group_meta(
                group_handler=group_handler,
                ngio_meta=meta,
            )
            self._group_handler = self._group_handler.reopen_handler()
            self._meta_handler = LabelsGroupMetaHandler(group_handler)

    @property
    def meta(self) -> NgioLabelsGroupMeta:
        """Return the metadata."""
        meta = self._meta_handler.get_meta()
        return meta

    def list(self) -> list[str]:
        """Create the /labels group if it doesn't exist."""
        return self.meta.labels

    def get(
        self,
        name: str,
        path: str | None = None,
        pixel_size: PixelSize | None = None,
        strict: bool = False,
    ) -> Label:
        """Get a label from the group.

        Args:
            name (str): The name of the label.
            path (str | None): The path to the image in the ome_zarr file.
            pixel_size (PixelSize | None): The pixel size of the image.
            strict (bool): Only used if the pixel size is provided. If True, the
                pixel size must match the image pixel size exactly. If False, the
                closest pixel size level will be returned.

        """
        if name not in self.list():
            raise NgioValueError(
                f"Label '{name}' not found in the Labels group. "
                f"Available labels: {self.list()}"
            )

        group_handler = self._group_handler.get_handler(name)
        label_meta_handler = LabelMetaHandler(group_handler)
        path = (
            label_meta_handler.get_meta()
            .get_dataset(path=path, pixel_size=pixel_size, strict=strict)
            .path
        )
        return Label(group_handler, path, label_meta_handler)

    def derive(
        self,
        name: str,
        ref_image: Image | Label,
        # Metadata parameters
        shape: Sequence[int] | None = None,
        pixelsize: float | tuple[float, float] | None = None,
        z_spacing: float | None = None,
        time_spacing: float | None = None,
        channels_policy: Literal["same", "squeeze", "singleton"] | int = "squeeze",
        ngff_version: NgffVersions | None = None,
        # Zarr Array parameters
        chunks: ChunksLike = "auto",
        shards: ShardsLike | None = None,
        dtype: str | None = None,
        dimension_separator: Literal[".", "/"] | None = None,
        compressors: CompressorLike | None = None,
        extra_array_kwargs: Mapping[str, Any] | None = None,
        overwrite: bool = False,
        # Deprecated arguments
        labels: Sequence[str] | None = None,
        pixel_size: PixelSize | None = None,
    ) -> "Label":
        """Create an empty OME-Zarr image from an existing image.

        If a kwarg is not provided, the value from the reference image will be used.

        Args:
            store (StoreOrGroup): The Zarr store or group to create the image in.
            ref_image (Image | Label): The reference image to derive the new image from.
            shape (Sequence[int] | None): The shape of the new image.
            pixelsize (float | tuple[float, float] | None): The pixel size of the new
                image.
            z_spacing (float | None): The z spacing of the new image.
            time_spacing (float | None): The time spacing of the new image.
            scaling_factors (Sequence[float] | Literal["auto"] | None): The scaling
                factors of the new image.
            axes_names (Sequence[str] | None): The axes names of the new image.
            name (str | None): The name of the new image.
            channels_meta (Sequence[str | Channel] | None): The channels metadata
                of the new image.
            channels_policy (Literal["squeeze", "same", "singleton"] | int):
                Possible policies:
                - If "squeeze", the channels axis will be removed (no matter its size).
                - If "same", the channels axis will be kept as is (if it exists).
                - If "singleton", the channels axis will be set to size 1.
                - If an integer is provided, the channels axis will be changed to have
                    that size.
            ngff_version (NgffVersions | None): The NGFF version to use.
            chunks (Sequence[int] | None): The chunk shape of the new image.
            shards (ShardsLike | None): The shard shape of the new image.
            dtype (str | None): The data type of the new image.
            dimension_separator (DIMENSION_SEPARATOR | None): The separator to use for
                dimensions.
            compressors (CompressorLike | None): The compressors to use.
            extra_array_kwargs (Mapping[str, Any] | None): Extra arguments to pass to
                the zarr array creation.
            overwrite (bool): Whether to overwrite an existing image.
            labels (Sequence[str] | None): The labels of the new image.
                This argument is deprecated please use channels_meta instead.
            pixel_size (PixelSize | None): The pixel size of the new image.
                This argument is deprecated please use pixelsize, z_spacing,
                and time_spacing instead.

        Returns:
            Label: The new derived label.

        """
        existing_labels = self.list()
        if name in existing_labels and not overwrite:
            raise NgioValueError(
                f"Label '{name}' already exists in the group. "
                "Use overwrite=True to replace it."
            )

        label_group = self._group_handler.get_group(name, create_mode=True)

        derive_label(
            ref_image=ref_image,
            store=label_group,
            shape=shape,
            pixelsize=pixelsize,
            z_spacing=z_spacing,
            time_spacing=time_spacing,
            name=name,
            channels_policy=channels_policy,
            ngff_version=ngff_version,
            chunks=chunks,
            shards=shards,
            dtype=dtype,
            dimension_separator=dimension_separator,
            compressors=compressors,
            extra_array_kwargs=extra_array_kwargs,
            overwrite=overwrite,
            labels=labels,
            pixel_size=pixel_size,
        )

        if name not in existing_labels:
            existing_labels.append(name)

        update_meta = NgioLabelsGroupMeta(
            labels=existing_labels, version=self.meta.version
        )
        self._meta_handler.update_meta(update_meta)
        return self.get(name)


def derive_label(
    *,
    store: StoreOrGroup,
    ref_image: Image | Label,
    # Metadata parameters
    shape: Sequence[int] | None = None,
    pixelsize: float | tuple[float, float] | None = None,
    z_spacing: float | None = None,
    time_spacing: float | None = None,
    name: str | None = None,
    channels_policy: Literal["same", "squeeze", "singleton"] | int = "squeeze",
    ngff_version: NgffVersions | None = None,
    # Zarr Array parameters
    chunks: ChunksLike = "auto",
    shards: ShardsLike | None = None,
    dtype: str | None = None,
    dimension_separator: Literal[".", "/"] | None = None,
    compressors: CompressorLike | None = None,
    extra_array_kwargs: Mapping[str, Any] | None = None,
    overwrite: bool = False,
    # Deprecated arguments
    labels: Sequence[str] | None = None,
    pixel_size: PixelSize | None = None,
) -> ZarrGroupHandler:
    """Derive a new OME-Zarr label from an existing image or label.

    If a kwarg is not provided, the value from the reference image will be used.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the label in.
        ref_image (Image | Label): The reference image to derive the new label from.
        shape (Sequence[int] | None): The shape of the new label.
        pixelsize (float | tuple[float, float] | None): The pixel size of the new label.
        z_spacing (float | None): The z spacing of the new label.
        time_spacing (float | None): The time spacing of the new label.
        name (str | None): The name of the new label.
        channels_policy (Literal["squeeze", "same", "singleton"] | int): Possible
            policies:
            - If "squeeze", the channels axis will be removed (no matter its size).
            - If "same", the channels axis will be kept as is (if it exists).
            - If "singleton", the channels axis will be set to size 1.
            - If an integer is provided, the channels axis will be changed to have that
                size.
        ngff_version (NgffVersions | None): The NGFF version to use.
        chunks (ChunksLike): The chunk shape of the new label. Defaults to "auto".
        shards (ShardsLike | None): The shard shape of the new label.
        dtype (str | None): The data type of the new label.
        dimension_separator (Literal[".", "/"] | None): The separator to use for
            dimensions.
        compressors (CompressorLike | None): The compressors to use.
        extra_array_kwargs (Mapping[str, Any] | None): Extra arguments to pass to
            the zarr array creation.
        overwrite (bool): Whether to overwrite an existing label. Defaults to False.
        labels (Sequence[str] | None): Deprecated. This argument is deprecated,
            please use channels_meta instead.
        pixel_size (PixelSize | None): Deprecated. The pixel size of the new label.
            This argument is deprecated, please use pixelsize, z_spacing,
            and time_spacing instead.

    Returns:
        ZarrGroupHandler: The group handler of the new label.

    """
    if dtype is None and isinstance(ref_image, Image):
        dtype = "uint32"
    group_handler = abstract_derive(
        ref_image=ref_image,
        meta_type=NgioLabelMeta,
        store=store,
        shape=shape,
        pixelsize=pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        name=name,
        channels_meta=None,
        channels_policy=channels_policy,
        ngff_version=ngff_version,
        chunks=chunks,
        shards=shards,
        dtype=dtype,
        dimension_separator=dimension_separator,
        compressors=compressors,
        extra_array_kwargs=extra_array_kwargs,
        overwrite=overwrite,
        labels=labels,
        pixel_size=pixel_size,
    )
    return group_handler


def build_masking_roi_table(label: Label) -> MaskingRoiTable:
    """Compute the masking ROI table for a label."""
    array = label.get_as_dask(axes_order=["t", "z", "y", "x"])
    rois = compute_masking_roi(array, label.pixel_size)
    return MaskingRoiTable(rois, reference_label=label.meta.name)



from pathlib import Path

import pandas as pd
import pytest

from ngio.common import Roi
from ngio.tables import RoiTable
from ngio.tables._tables_container import open_table, write_table
from ngio.tables.backends import AnnDataBackend
from ngio.tables.v1 import RoiTableV1
from ngio.tables.v1._roi_table import RoiTableV1Meta
from ngio.utils import NgioValueError, ZarrGroupHandler


def test_roi_table_v1(tmp_path: Path):
    rois = [
        Roi.from_values(
            name="roi1",
            slices={"x": slice(0, 1), "y": slice(0, 1), "z": slice(0, 1)},
        )
    ]

    table = RoiTableV1(rois=rois)
    assert isinstance(table.__repr__(), str)

    table.add(
        roi=Roi.from_values(
            name="roi2",
            slices={"x": slice(0, 1), "y": slice(0, 1), "z": slice(0, 1)},
        )
    )

    with pytest.raises(NgioValueError):
        # ROI name already exists
        table.add(
            roi=Roi.from_values(
                name="roi2",
                slices={"x": slice(0, 1), "y": slice(0, 1), "z": slice(0, 1)},
            )
        )

    table.add(
        roi=Roi.from_values(
            name="roi2",
            slices={"x": slice(0, 1), "y": slice(0, 1), "z": slice(0, 1)},
        ),
        overwrite=True,
    )
    assert len(table.rois()) == 2
    write_table(store=tmp_path / "roi_table.zarr", table=table, backend="anndata")

    loaded_table = open_table(store=tmp_path / "roi_table.zarr")
    assert isinstance(loaded_table, RoiTableV1)
    assert len(loaded_table.rois()) == 2
    assert loaded_table.get("roi1") == table.get("roi1")
    assert loaded_table.get("roi2") == table.get("roi2")

    with pytest.raises(NgioValueError):
        loaded_table.get("roi3")

    assert loaded_table.meta.backend == "anndata"
    meta_dict = loaded_table._meta.model_dump()
    assert meta_dict.get("table_version") == loaded_table.version()
    assert meta_dict.get("type") == loaded_table.table_type()


def test_roi_no_index(tmp_path: Path):
    """ngio needs to support reading a table without an index. for legacy reasons"""
    handler = ZarrGroupHandler(tmp_path / "roi_table.zarr")
    backend = AnnDataBackend()
    backend.set_group_handler(handler)

    roi_table = pd.DataFrame(
        {
            "x_micrometer": [0.0, 1.0],
            "y_micrometer": [0.0, 1.0],
            "z_micrometer": [0.0, 1.0],
            "len_x_micrometer": [1.0, 1.0],
            "len_y_micrometer": [1.0, 1.0],
            "len_z_micrometer": [1.0, 1.0],
        }
    )
    roi_table.index = pd.Index(["roi_1", "roi_2"])

    backend.write(
        roi_table,
        metadata=RoiTableV1Meta().model_dump(exclude_none=True),
    )

    roi_table = RoiTable.from_handler(handler=handler)
    assert isinstance(roi_table, RoiTable)
    assert len(roi_table.rois()) == 2



"""Aggregation and filtering operations for tables."""

import asyncio
from collections import Counter
from collections.abc import Sequence
from dataclasses import dataclass, field
from typing import Literal

import pandas as pd
import polars as pl

from ngio.images._ome_zarr_container import OmeZarrContainer
from ngio.tables import Table, TableType


@dataclass
class TableWithExtras:
    """A class to hold a table and its extras."""

    table: Table
    extras: dict[str, str] = field(default_factory=dict)


def _reindex_dataframe(
    dataframe, index_cols: list[str], index_key: str | None = None
) -> pd.DataFrame:
    """Reindex a dataframe using an hash of the index columns."""
    # Reindex the dataframe
    old_index = dataframe.index.name
    if old_index is not None:
        dataframe = dataframe.reset_index()
        index_cols.append(old_index)
    dataframe.index = dataframe[index_cols].astype(str).agg("_".join, axis=1)

    if index_key is None:
        dataframe.index.name = index_key
    return dataframe


def _add_const_columns(
    dataframe: pd.DataFrame,
    new_cols: dict[str, str],
    index_key: str | None = None,
) -> pd.DataFrame:
    for col, value in new_cols.items():
        dataframe[col] = value

    if index_key is not None:
        dataframe = _reindex_dataframe(
            dataframe=dataframe,
            index_cols=list(new_cols.keys()),
            index_key=index_key,
        )
    return dataframe


def _add_const_columns_pl(
    dataframe: pl.LazyFrame,
    new_cols: dict[str, str],
    index_key: str | None = None,
) -> pl.LazyFrame:
    dataframe = dataframe.with_columns(
        [pl.lit(value, dtype=pl.String()).alias(col) for col, value in new_cols.items()]
    )

    if index_key is not None:
        dataframe = dataframe.with_columns(
            [
                pl.concat_str(
                    [pl.col(col) for col in new_cols.keys()],
                    separator="_",
                ).alias(index_key)
            ]
        )
    return dataframe


def _pd_concat(
    tables: Sequence[TableWithExtras], index_key: str | None = None
) -> pd.DataFrame:
    """Concatenate tables from different plates into a single table."""
    if len(tables) == 0:
        raise ValueError("No tables to concatenate.")

    dataframes = []
    for table in tables:
        dataframe = _add_const_columns(
            dataframe=table.table.dataframe, new_cols=table.extras, index_key=index_key
        )
        dataframes.append(dataframe)
    concatenated_table = pd.concat(dataframes, axis=0)
    return concatenated_table


def _pl_concat(
    tables: Sequence[TableWithExtras], index_key: str | None = None
) -> pl.LazyFrame:
    """Concatenate tables from different plates into a single table."""
    if len(tables) == 0:
        raise ValueError("No tables to concatenate.")

    dataframes = []
    for table in tables:
        polars_ls = _add_const_columns_pl(
            dataframe=table.table.lazy_frame,
            new_cols=table.extras,
            index_key=index_key,
        )
        dataframes.append(polars_ls)

    concatenated_table = pl.concat(dataframes, how="vertical")
    return concatenated_table


def conctatenate_tables(
    tables: Sequence[TableWithExtras],
    mode: Literal["eager", "lazy"] = "eager",
    index_key: str | None = None,
    table_cls: type[TableType] | None = None,
) -> Table:
    """Concatenate tables from different plates into a single table."""
    if len(tables) == 0:
        raise ValueError("No tables to concatenate.")

    table0 = next(iter(tables)).table

    if mode == "lazy":
        concatenated_table = _pl_concat(tables=tables, index_key=index_key)
    elif mode == "eager":
        concatenated_table = _pd_concat(tables=tables, index_key=index_key)
    else:
        raise ValueError(f"Unknown mode: {mode}. Use 'eager' or 'lazy'.")

    meta = table0.meta
    meta.index_key = index_key
    meta.index_type = "str"

    if table_cls is not None:
        return table_cls.from_table_data(
            table_data=concatenated_table,
            meta=meta,
        )
    return table0.from_table_data(
        table_data=concatenated_table,
        meta=meta,
    )


def _check_images_and_extras(
    images: Sequence[OmeZarrContainer],
    extras: Sequence[dict[str, str]],
) -> None:
    """Check if the images and extras are valid."""
    if len(images) == 0:
        raise ValueError("No images to concatenate.")

    if len(images) != len(extras):
        raise ValueError("The number of images and extras must be the same.")


def _concatenate_image_tables(
    images: Sequence[OmeZarrContainer],
    extras: Sequence[dict[str, str]],
    name: str,
    table_cls: type[TableType] | None = None,
    index_key: str | None = None,
    strict: bool = True,
    mode: Literal["eager", "lazy"] = "eager",
) -> Table:
    """Concatenate tables from different images into a single table."""
    _check_images_and_extras(images=images, extras=extras)

    tables = []
    for image, extra in zip(images, extras, strict=True):
        if not strict and name not in image.list_tables():
            continue
        table = image.get_table(name)
        tables.append(TableWithExtras(table=table, extras=extra))

    return conctatenate_tables(
        tables=tables,
        mode=mode,
        index_key=index_key,
        table_cls=table_cls,
    )


def concatenate_image_tables(
    images: Sequence[OmeZarrContainer],
    extras: Sequence[dict[str, str]],
    name: str,
    index_key: str | None = None,
    strict: bool = True,
    mode: Literal["eager", "lazy"] = "eager",
) -> Table:
    """Concatenate tables from different images into a single table.

    Args:
        images: A collection of images.
        extras: A collection of extras dictionaries for each image.
            this will be added as columns to the table, and will be
            concatenated with the table index to create a new index.
        name: The name of the table to concatenate.
        index_key: The key to use for the index of the concatenated table.
        strict: If True, raise an error if the table is not found in the image.
        mode: The mode to use for concatenation. Can be 'eager' or 'lazy'.
            if 'eager', the table will be loaded into memory.
            if 'lazy', the table will be loaded as a lazy frame.
    """
    return _concatenate_image_tables(
        images=images,
        extras=extras,
        name=name,
        table_cls=None,
        index_key=index_key,
        strict=strict,
        mode=mode,
    )


def concatenate_image_tables_as(
    images: Sequence[OmeZarrContainer],
    extras: Sequence[dict[str, str]],
    name: str,
    table_cls: type[TableType],
    index_key: str | None = None,
    strict: bool = True,
    mode: Literal["eager", "lazy"] = "eager",
) -> TableType:
    """Concatenate tables from different images into a single table.

    Args:
        images: A collection of images.
        extras: A collection of extras dictionaries for each image.
            this will be added as columns to the table, and will be
            concatenated with the table index to create a new index.
        name: The name of the table to concatenate.
        table_cls: The output will be casted to this class, if the new table_cls is
            compatible with the table_cls of the input tables.
        index_key: The key to use for the index of the concatenated table.
        strict: If True, raise an error if the table is not found in the image.
        mode: The mode to use for concatenation. Can be 'eager' or 'lazy'.
            if 'eager', the table will be loaded into memory.
            if 'lazy', the table will be loaded as a lazy frame.
    """
    table = _concatenate_image_tables(
        images=images,
        extras=extras,
        name=name,
        table_cls=table_cls,
        index_key=index_key,
        strict=strict,
        mode=mode,
    )
    if not isinstance(table, table_cls):
        raise ValueError(f"Table is not of type {table_cls}. Got {type(table)}")
    return table


async def _concatenate_image_tables_async(
    images: Sequence[OmeZarrContainer],
    extras: Sequence[dict[str, str]],
    name: str,
    table_cls: type[TableType] | None = None,
    index_key: str | None = None,
    strict: bool = True,
    mode: Literal["eager", "lazy"] = "eager",
) -> Table:
    """Concatenate tables from different images into a single table."""
    _check_images_and_extras(images=images, extras=extras)

    def process_image(
        image: OmeZarrContainer,
        name: str,
        extra: dict[str, str],
        mode: Literal["eager", "lazy"] = "eager",
        strict: bool = True,
    ) -> TableWithExtras | None:
        """Process a single image and return the table."""
        if not strict and name not in image.list_tables():
            return None
        _table = image.get_table(name)
        if mode == "lazy":
            # make sure the table is loaded lazily
            # It the backend is not lazy, this will be
            # loaded eagerly
            _ = _table.lazy_frame
        elif mode == "eager":
            # make sure the table is loaded eagerly
            _ = _table.dataframe
        table = TableWithExtras(
            table=_table,
            extras=extra,
        )
        return table

    tasks = []
    for image, extra in zip(images, extras, strict=True):
        task = asyncio.to_thread(
            process_image,
            image=image,
            name=name,
            extra=extra,
            strict=strict,
        )
        tasks.append(task)
    tables = await asyncio.gather(*tasks)
    tables = [table for table in tables if table is not None]
    return conctatenate_tables(
        tables=tables,
        mode=mode,
        index_key=index_key,
        table_cls=table_cls,
    )


async def concatenate_image_tables_async(
    images: Sequence[OmeZarrContainer],
    extras: Sequence[dict[str, str]],
    name: str,
    index_key: str | None = None,
    strict: bool = True,
    mode: Literal["eager", "lazy"] = "eager",
) -> Table:
    """Concatenate tables from different images into a single table.

    Args:
        images: A collection of images.
        extras: A collection of extras dictionaries for each image.
            this will be added as columns to the table, and will be
            concatenated with the table index to create a new index.
        name: The name of the table to concatenate.
        index_key: The key to use for the index of the concatenated table.
        strict: If True, raise an error if the table is not found in the image.
        mode: The mode to use for concatenation. Can be 'eager' or 'lazy'.
            if 'eager', the table will be loaded into memory.
            if 'lazy', the table will be loaded as a lazy frame.
    """
    return await _concatenate_image_tables_async(
        images=images,
        extras=extras,
        name=name,
        table_cls=None,
        index_key=index_key,
        strict=strict,
        mode=mode,
    )


async def concatenate_image_tables_as_async(
    images: Sequence[OmeZarrContainer],
    extras: Sequence[dict[str, str]],
    name: str,
    table_cls: type[TableType],
    index_key: str | None = None,
    strict: bool = True,
    mode: Literal["eager", "lazy"] = "eager",
) -> TableType:
    """Concatenate tables from different images into a single table.

    Args:
        images: A collection of images.
        extras: A collection of extras dictionaries for each image.
            this will be added as columns to the table, and will be
            concatenated with the table index to create a new index.
        name: The name of the table to concatenate.
        table_cls: The output will be casted to this class, if the new table_cls is
            compatible with the table_cls of the input tables.
        index_key: The key to use for the index of the concatenated table.
        strict: If True, raise an error if the table is not found in the image.
        mode: The mode to use for concatenation. Can be 'eager' or 'lazy'.
            if 'eager', the table will be loaded into memory.
            if 'lazy', the table will be loaded as a lazy frame.
    """
    table = await _concatenate_image_tables_async(
        images=images,
        extras=extras,
        name=name,
        table_cls=table_cls,
        index_key=index_key,
        strict=strict,
        mode=mode,
    )
    if not isinstance(table, table_cls):
        raise ValueError(f"Table is not of type {table_cls}. Got {type(table)}")
    return table


def _tables_names_coalesce(
    tables_names: list[list[str]],
    mode: Literal["common", "all"] = "common",
) -> list[str]:
    num_images = len(tables_names)
    if num_images == 0:
        raise ValueError("No images to concatenate.")

    names = [name for _names in tables_names for name in _names]
    names_counts = Counter(names)

    if mode == "common":
        # Get the names that are present in all images
        common_names = [
            name for name, count in names_counts.items() if count == num_images
        ]
        return common_names
    elif mode == "all":
        # Get all names
        return list(names_counts.keys())
    else:
        raise ValueError(f"Unknown mode: {mode}. Use 'common' or 'all'.")


def list_image_tables(
    images: Sequence[OmeZarrContainer],
    filter_types: str | None = None,
    mode: Literal["common", "all"] = "common",
) -> list[str]:
    """List all table names in the images.

    Args:
        images: A collection of images.
        filter_types (str | None): The type of tables to filter. If None,
            return all tables. Defaults to None.
        mode (Literal["common", "all"]): Whether to return only common tables
            between all images or all tables. Defaults to "common".
    """
    tables_names = []
    for image in images:
        tables = image.list_tables(filter_types=filter_types)
        tables_names.append(tables)

    return _tables_names_coalesce(
        tables_names=tables_names,
        mode=mode,
    )


async def list_image_tables_async(
    images: Sequence[OmeZarrContainer],
    filter_types: str | None = None,
    mode: Literal["common", "all"] = "common",
) -> list[str]:
    """List all image tables in the image asynchronously.

    Args:
        images: A collection of images.
        filter_types (str | None): The type of tables to filter. If None,
            return all tables. Defaults to None.
        mode (Literal["common", "all"]): Whether to return only common tables
            between all images or all tables. Defaults to "common".
    """
    images_ids = []

    # key table name, value list of paths
    def process_image(
        image: OmeZarrContainer, filter_types: str | None = None
    ) -> list[str]:
        tables = image.list_tables(filter_types=filter_types)
        return tables

    tasks = []
    for i, image in enumerate(images):
        images_ids.append(i)
        task = asyncio.to_thread(process_image, image, filter_types=filter_types)
        tasks.append(task)

    tables_names = await asyncio.gather(*tasks)
    return _tables_names_coalesce(
        tables_names=tables_names,
        mode=mode,
    )



"""Utility functions for working with OME-Zarr images."""

import warnings
from collections.abc import Mapping, Sequence
from typing import Any, Literal, TypeVar

from zarr.core.array import CompressorLike

from ngio.common._pyramid import ChunksLike, ImagePyramidBuilder, ShardsLike
from ngio.ome_zarr_meta import (
    NgioImageMeta,
    NgioLabelMeta,
    update_ngio_meta,
)
from ngio.ome_zarr_meta.ngio_specs import (
    AxesHandler,
    Channel,
    ChannelsMeta,
    DefaultNgffVersion,
    DefaultSpaceUnit,
    DefaultTimeUnit,
    NgffVersions,
    SpaceUnits,
    TimeUnits,
    build_canonical_axes_handler,
    canonical_axes_order,
    canonical_label_axes_order,
)
from ngio.ome_zarr_meta.ngio_specs._axes import AxesSetup
from ngio.utils import NgioValueError, StoreOrGroup, ZarrGroupHandler

_image_or_label_meta = TypeVar("_image_or_label_meta", NgioImageMeta, NgioLabelMeta)


def _build_axes_handler(
    *,
    shape: tuple[int, ...],
    axes_names: Sequence[str] | None,
    default_channel_order: tuple[str, ...],
    space_units: SpaceUnits | str | None = DefaultSpaceUnit,
    time_units: TimeUnits | str | None = DefaultTimeUnit,
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = False,
) -> AxesHandler:
    """Compute axes names for given shape."""
    if axes_names is None:
        axes_names = default_channel_order[-len(shape) :]
    # Validate length
    if len(axes_names) != len(shape):
        raise NgioValueError(
            f"Number of axes names {axes_names} does not match the number of "
            f"dimensions {shape}."
        )
    return build_canonical_axes_handler(
        axes_names=axes_names,
        space_units=space_units,
        time_units=time_units,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )


def _align_to_axes(
    *,
    values: dict[str, float],
    axes_handler: AxesHandler,
    default_value: float = 1.0,
) -> tuple[float, ...]:
    """Align given values to axes names."""
    aligned_values = [default_value] * len(axes_handler.axes_names)
    for ax, value in values.items():
        index = axes_handler.get_index(ax)
        if index is not None:
            aligned_values[index] = value
    return tuple(aligned_values)


def _check_deprecated_scaling_factors(
    *,
    yx_scaling_factor: float | tuple[float, float] | None = None,
    z_scaling_factor: float | None = None,
    scaling_factors: Sequence[float] | Literal["auto"] = "auto",
    shape: tuple[int, ...],
) -> Sequence[float] | Literal["auto"]:
    if yx_scaling_factor is not None or z_scaling_factor is not None:
        warnings.warn(
            "The 'yx_scaling_factor' and 'z_scaling_factor' arguments are deprecated "
            "and will be removed in future versions. Please use the 'scaling_factors' "
            "argument instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if scaling_factors != "auto":
            raise NgioValueError(
                "Cannot use both 'scaling_factors' and deprecated "
                "'yx_scaling_factor'/'z_scaling_factor' arguments."
            )
        if isinstance(yx_scaling_factor, tuple):
            if len(yx_scaling_factor) != 2:
                raise NgioValueError(
                    "yx_scaling_factor tuple must have length 2 for y and x scaling."
                )
            y_scale = yx_scaling_factor[0]
            x_scale = yx_scaling_factor[1]
        else:
            y_scale = yx_scaling_factor if yx_scaling_factor is not None else 2.0
            x_scale = yx_scaling_factor if yx_scaling_factor is not None else 2.0
        z_scale = z_scaling_factor if z_scaling_factor is not None else 1.0
        scaling_factors = (z_scale, x_scale, y_scale)
        if len(scaling_factors) < len(shape):
            padding = (1.0,) * (len(shape) - len(scaling_factors))
            scaling_factors = padding + scaling_factors

        return scaling_factors
    return scaling_factors


def _compute_scaling_factors(
    *,
    scaling_factors: Sequence[float] | Literal["auto"],
    shape: tuple[int, ...],
    axes_handler: AxesHandler,
    xy_scaling_factor: float | tuple[float, float] | None = None,
    z_scaling_factor: float | None = None,
) -> tuple[float, ...]:
    """Compute scaling factors for given axes names."""
    # TODO remove with ngio 0.6
    scaling_factors = _check_deprecated_scaling_factors(
        yx_scaling_factor=xy_scaling_factor,
        z_scaling_factor=z_scaling_factor,
        scaling_factors=scaling_factors,
        shape=shape,
    )
    if scaling_factors == "auto":
        return _align_to_axes(
            values={
                "x": 2.0,
                "y": 2.0,
                "z": 1.0,
            },
            axes_handler=axes_handler,
        )
    if len(scaling_factors) != len(shape):
        raise NgioValueError(
            "Length of scaling_factors does not match the number of dimensions."
        )
    return tuple(scaling_factors)


def _compute_base_scale(
    *,
    pixelsize: float | tuple[float, float],
    z_spacing: float,
    time_spacing: float,
    axes_handler: AxesHandler,
) -> tuple[float, ...]:
    """Compute base scale for given axes names."""
    if isinstance(pixelsize, tuple):
        if len(pixelsize) != 2:
            raise NgioValueError(
                "pixelsize tuple must have length 2 for y and x pixel sizes."
            )
        x_size = pixelsize[1]
        y_size = pixelsize[0]
    else:
        x_size = pixelsize
        y_size = pixelsize
    return _align_to_axes(
        values={
            "x": x_size,
            "y": y_size,
            "z": z_spacing,
            "t": time_spacing,
        },
        axes_handler=axes_handler,
    )


def _create_image_like_group(
    *,
    store: StoreOrGroup,
    pyramid_builder: ImagePyramidBuilder,
    meta: _image_or_label_meta,
    overwrite: bool = False,
) -> ZarrGroupHandler:
    """Advanced create empty image container function placeholder."""
    mode = "w" if overwrite else "w-"
    group_handler = ZarrGroupHandler(
        store=store, mode=mode, cache=False, zarr_format=meta.zarr_format
    )
    update_ngio_meta(group_handler, meta)
    # Reopen in r+ mode
    group_handler = group_handler.reopen_handler()
    # Write the pyramid
    pyramid_builder.to_zarr(group=group_handler.group)
    return group_handler


def _add_channels_meta(
    *,
    meta: _image_or_label_meta,
    channels_meta: Sequence[str | Channel] | None = None,
) -> _image_or_label_meta:
    """Create ChannelsMeta from given channels_meta input."""
    if isinstance(meta, NgioLabelMeta):
        if channels_meta is not None:
            raise NgioValueError(
                "Cannot add channels_meta to NgioLabelMeta. "
                "Labels do not have channels."
            )
        else:
            return meta
    if channels_meta is None:
        return meta
    list_of_channels = []
    for c in channels_meta:
        if isinstance(c, str):
            channel = Channel.default_init(label=c)
        elif isinstance(c, Channel):
            channel = c
        else:
            raise NgioValueError(
                "channels_meta must be a list of strings or Channel objects."
            )
        list_of_channels.append(channel)

    channels_meta_ = ChannelsMeta(channels=list_of_channels)
    meta.set_channels_meta(channels_meta=channels_meta_)
    return meta


def init_image_like(
    *,
    # Where to create the image
    store: StoreOrGroup,
    # Ngff image parameters
    meta_type: type[_image_or_label_meta],
    shape: Sequence[int],
    pixelsize: float | tuple[float, float],
    z_spacing: float = 1.0,
    time_spacing: float = 1.0,
    scaling_factors: Sequence[float] | Literal["auto"] = "auto",
    levels: int | list[str] = 5,
    space_unit: SpaceUnits | str | None = DefaultSpaceUnit,
    time_unit: TimeUnits | str | None = DefaultTimeUnit,
    axes_names: Sequence[str] | None = None,
    name: str | None = None,
    channels_meta: Sequence[str | Channel] | None = None,
    ngff_version: NgffVersions = DefaultNgffVersion,
    # Zarr Array parameters
    chunks: ChunksLike = "auto",
    shards: ShardsLike | None = None,
    dtype: str = "uint16",
    dimension_separator: Literal[".", "/"] = "/",
    compressors: CompressorLike = "auto",
    extra_array_kwargs: Mapping[str, Any] | None = None,
    # internal axes configuration for advanced use cases
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = False,
    # Whether to overwrite existing image
    overwrite: bool = False,
    # Deprecated arguments
    yx_scaling_factor: float | tuple[float, float] | None = None,
    z_scaling_factor: float | None = None,
) -> ZarrGroupHandler:
    """Create an empty OME-Zarr image with the given shape and metadata."""
    shape = tuple(shape)
    if meta_type is NgioImageMeta:
        default_axes_order = canonical_axes_order()
    else:
        default_axes_order = canonical_label_axes_order()

    axes_handler = _build_axes_handler(
        shape=shape,
        axes_names=axes_names,
        default_channel_order=default_axes_order,
        space_units=space_unit,
        time_units=time_unit,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )
    base_scale = _compute_base_scale(
        pixelsize=pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        axes_handler=axes_handler,
    )
    scaling_factors = _compute_scaling_factors(
        scaling_factors=scaling_factors,
        shape=shape,
        axes_handler=axes_handler,
        xy_scaling_factor=yx_scaling_factor,
        z_scaling_factor=z_scaling_factor,
    )
    if isinstance(levels, int):
        levels_paths = tuple(str(i) for i in range(levels))
    else:
        levels_paths = tuple(levels)

    pyramid_builder = ImagePyramidBuilder.from_scaling_factors(
        levels_paths=levels_paths,
        scaling_factors=scaling_factors,
        base_shape=shape,
        base_scale=base_scale,
        axes=axes_handler.axes_names,
        chunks=chunks,
        data_type=dtype,
        dimension_separator=dimension_separator,
        compressors=compressors,
        shards=shards,
        zarr_format=2 if ngff_version == "0.4" else 3,
        other_array_kwargs=extra_array_kwargs,
    )
    meta = meta_type.default_init(
        levels=[p.path for p in pyramid_builder.levels],
        axes_handler=axes_handler,
        scales=[p.scale for p in pyramid_builder.levels],
        translations=[None for _ in pyramid_builder.levels],
        name=name,
        version=ngff_version,
    )
    meta = _add_channels_meta(meta=meta, channels_meta=channels_meta)
    # Keep this creation at the end to avoid partial creations on errors
    return _create_image_like_group(
        store=store,
        pyramid_builder=pyramid_builder,
        meta=meta,
        overwrite=overwrite,
    )


def init_image_like_from_shapes(
    *,
    # Where to create the image
    store: StoreOrGroup,
    # Ngff image parameters
    meta_type: type[_image_or_label_meta],
    shapes: Sequence[tuple[int, ...]],
    pixelsize: float | tuple[float, float],
    z_spacing: float = 1.0,
    time_spacing: float = 1.0,
    levels: list[str] | None = None,
    space_unit: SpaceUnits | str | None = DefaultSpaceUnit,
    time_unit: TimeUnits | str | None = DefaultTimeUnit,
    axes_names: Sequence[str] | None = None,
    name: str | None = None,
    channels_meta: Sequence[str | Channel] | None = None,
    ngff_version: NgffVersions = DefaultNgffVersion,
    # Zarr Array parameters
    chunks: ChunksLike = "auto",
    shards: ShardsLike | None = None,
    dtype: str = "uint16",
    dimension_separator: Literal[".", "/"] = "/",
    compressors: CompressorLike = "auto",
    extra_array_kwargs: Mapping[str, Any] | None = None,
    # internal axes configuration for advanced use cases
    axes_setup: AxesSetup | None = None,
    allow_non_canonical_axes: bool = False,
    strict_canonical_order: bool = False,
    # Whether to overwrite existing image
    overwrite: bool = False,
) -> ZarrGroupHandler:
    """Create an empty OME-Zarr image with the given shape and metadata."""
    base_shape = shapes[0]
    if meta_type is NgioImageMeta:
        default_axes_order = canonical_axes_order()
    else:
        default_axes_order = canonical_label_axes_order()

    axes_handler = _build_axes_handler(
        shape=base_shape,
        axes_names=axes_names,
        default_channel_order=default_axes_order,
        space_units=space_unit,
        time_units=time_unit,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )
    base_scale = _compute_base_scale(
        pixelsize=pixelsize,
        z_spacing=z_spacing,
        time_spacing=time_spacing,
        axes_handler=axes_handler,
    )
    if levels is None:
        levels_paths = tuple(str(i) for i in range(len(shapes)))
    else:
        levels_paths = tuple(levels)

    pyramid_builder = ImagePyramidBuilder.from_shapes(
        shapes=shapes,
        base_scale=base_scale,
        levels_paths=levels_paths,
        axes=axes_handler.axes_names,
        chunks=chunks,
        data_type=dtype,
        dimension_separator=dimension_separator,
        compressors=compressors,
        shards=shards,
        zarr_format=2 if ngff_version == "0.4" else 3,
        other_array_kwargs=extra_array_kwargs,
    )
    meta = meta_type.default_init(
        levels=[p.path for p in pyramid_builder.levels],
        axes_handler=axes_handler,
        scales=[p.scale for p in pyramid_builder.levels],
        translations=[None for _ in pyramid_builder.levels],
        name=name,
        version=ngff_version,
    )
    meta = _add_channels_meta(meta=meta, channels_meta=channels_meta)
    # Keep this creation at the end to avoid partial creations on errors
    return _create_image_like_group(
        store=store,
        pyramid_builder=pyramid_builder,
        meta=meta,
        overwrite=overwrite,
    )



"""Implementation of a generic table class."""

from ngio.tables._abstract_table import AbstractBaseTable
from ngio.tables.backends import BackendMeta, TableBackend
from ngio.utils import ZarrGroupHandler


class GenericTableMeta(BackendMeta):
    """Metadata for the generic table.

    This is used to store metadata for a generic table.
    It does not have a specific definition.
    """

    table_version: str | None = "1"
    type: str | None = "generic_table"


class GenericTable(AbstractBaseTable):
    """Class to a non-specific table.

    This can be used to load any table that does not have
    a specific definition.
    """

    @staticmethod
    def table_type() -> str:
        """Return the type of the table."""
        return "generic_table"

    @staticmethod
    def version() -> str:
        """The generic table does not have a version.

        Since does not follow a specific schema.
        """
        return "1"

    @classmethod
    def from_handler(
        cls,
        handler: ZarrGroupHandler,
        backend: TableBackend | None = None,
    ) -> "GenericTable":
        return cls._from_handler(
            handler=handler,
            backend=backend,
            meta_model=BackendMeta,
        )



import pytest

from ngio import PixelSize
from ngio.common import Roi
from ngio.utils import NgioValueError


def test_basic_rois_ops():
    roi = Roi.from_values(
        name="test",
        slices={
            "x": (0, 1),
            "y": (0, 1),
            "z": (0, 1),
        },
        space="world",
        other="other",  # type: ignore
        label=1,
    )

    slixe_x = roi.get("x")
    assert slixe_x is not None
    assert slixe_x.axis_name == "x"
    assert slixe_x.start == 0
    assert slixe_x.length == 1

    pixel_size = PixelSize(x=1.0, y=1.0, z=1.0)
    raster_roi = roi.to_pixel(pixel_size)
    assert roi.__str__()
    assert roi.__repr__()

    assert raster_roi.to_slicing_dict(pixel_size=pixel_size) == {
        "x": slice(0.0, 1.0),
        "y": slice(0.0, 1.0),
        "z": slice(0.0, 1.0),
    }
    assert roi.model_extra is not None
    assert roi.model_extra["other"] == "other"

    world_roi_2 = raster_roi.to_world(pixel_size)

    x_slice_2 = world_roi_2.get("x")
    assert x_slice_2 is not None
    assert x_slice_2.axis_name == "x"
    assert x_slice_2.start == 0
    assert x_slice_2.length == 1

    y_slice_2 = world_roi_2.get("y")
    assert y_slice_2 is not None
    assert y_slice_2.axis_name == "y"
    assert y_slice_2.start == 0
    assert y_slice_2.length == 1
    assert world_roi_2.other == "other"  # type: ignore

    roi_zoomed = roi.zoom(2.0)
    with pytest.raises(ValueError):
        roi.zoom(-1.0)

    assert roi_zoomed.to_slicing_dict(pixel_size) == {
        "x": slice(0.0, 2.0),
        "y": slice(0.0, 2.0),
        "z": slice(0.0, 1.0),
    }

    roi2 = Roi.from_values(
        name="test2",
        slices={
            "x": (0.0, 1.0),
            "y": (0.0, 1.0),
            "z": (0.0, 1.0),
        },
        space="world",
        # type: ignore
        label=1,
    )
    roi_i = roi.intersection(roi2)
    assert roi_i is not None
    assert roi_i.label == 1

    roi2.label = 2
    with pytest.raises(NgioValueError):
        roi.intersection(roi2)


@pytest.mark.parametrize(
    "roi_ref,roi_other,expected_intersection,expected_name",
    [
        (
            # Basic intersection
            Roi.from_values(
                name="ref",
                slices={
                    "x": (0.0, 1.0),
                    "y": (0.0, 1.0),
                    "z": (0.0, 1.0),
                },
                space="world",
            ),
            Roi.from_values(
                name="other",
                slices={
                    "x": (0.5, 1.0),
                    "y": (0.5, 1.0),
                    "z": (0.5, 1.0),
                },
                space="world",
            ),
            Roi.from_values(
                name="ref:other",
                slices={
                    "x": (0.5, 0.5),
                    "y": (0.5, 0.5),
                    "z": (0.5, 0.5),
                },
                space="world",
            ),
            "ref:other",
        ),
        (
            # No intersection
            Roi.from_values(
                name="ref",
                slices={
                    "x": (0.0, 1.0),
                    "y": (0.0, 1.0),
                    "z": (0.0, 1.0),
                },
                space="world",
            ),
            Roi.from_values(
                name="other",
                slices={
                    "x": (2.0, 1.0),
                    "y": (2.0, 1.0),
                    "z": (2.0, 1.0),
                },
                space="world",
            ),
            None,
            "",
        ),
        (
            # Intersection with z=None (expected behaves like infinite z)
            # t=None (expected behaves like infinite t)
            Roi.from_values(
                name="ref",
                slices={
                    "x": (0.0, 1.0),
                    "y": (0.0, 1.0),
                },
                space="world",
            ),
            Roi.from_values(
                name=None,
                slices={
                    "x": (0.5, 1.0),
                    "y": (0.5, 1.0),
                    "z": (-1.0, 2.0),
                    "t": (0.0, 2.0),
                },
                space="world",
                unit="micrometer",
            ),
            Roi.from_values(
                name="ref",
                slices={
                    "x": (0.5, 0.5),
                    "y": (0.5, 0.5),
                    "z": (-1.0, 2.0),
                    "t": (0.0, 2.0),
                },
                space="world",
            ),
            "ref",
        ),
    ],
)
def test_rois_intersection(
    roi_ref: Roi,
    roi_other: Roi,
    expected_intersection: Roi | None,
    expected_name: str,
):
    intersection = roi_ref.intersection(roi_other)
    if expected_intersection is None:
        assert intersection is None
    else:
        assert intersection is not None
        assert intersection.name == expected_name
        assert intersection.get("x") == expected_intersection.get("x")
        assert intersection.get("y") == expected_intersection.get("y")
        assert intersection.get("z") == expected_intersection.get("z")

        assert intersection.get("t") == expected_intersection.get("t")


@pytest.mark.parametrize(
    "roi_ref,roi_other,expected_union,expected_name",
    [
        (
            # Basic intersection
            Roi.from_values(
                name="ref",
                slices={
                    "x": (0.0, 1.0),
                    "y": (0.0, 1.0),
                    "z": (0.0, 1.0),
                },
                space="world",
            ),
            Roi.from_values(
                name="other",
                slices={
                    "x": (0.5, 1.0),
                    "y": (0.5, 1.0),
                    "z": (0.5, 1.0),
                },
                space="world",
            ),
            Roi.from_values(
                name="ref:other",
                slices={
                    "x": (0.0, 1.5),
                    "y": (0.0, 1.5),
                    "z": (0.0, 1.5),
                },
                space="world",
            ),
            "ref:other",
        ),
        (
            # Intersection with z=None (expected behaves like infinite z)
            # t=None (expected behaves like infinite t)
            Roi.from_values(
                name="ref",
                slices={
                    "x": (0.0, 1.0),
                    "y": (0.0, 1.0),
                },
                space="world",
            ),
            Roi.from_values(
                name=None,
                slices={
                    "x": (0.5, 1.0),
                    "y": (0.5, 1.0),
                    "z": (-1.0, 2.0),
                    "t": (0.0, 2.0),
                },
                space="world",
                unit="micrometer",
            ),
            Roi.from_values(
                name="ref",
                slices={
                    "x": (0.0, 1.5),
                    "y": (0.0, 1.5),
                    "z": (-1.0, 2.0),
                    "t": (0.0, 2.0),
                },
                space="world",
            ),
            "ref",
        ),
    ],
)
def test_rois_union(
    roi_ref: Roi,
    roi_other: Roi,
    expected_union: Roi,
    expected_name: str,
):
    union = roi_ref.union(roi_other)
    assert union is not None
    assert union.name == expected_name
    assert union.get("x") == expected_union.get("x")
    assert union.get("y") == expected_union.get("y")
    assert union.get("z") == expected_union.get("z")
    assert union.get("t") == expected_union.get("t")



"""Dimension metadata.

This is not related to the NGFF metadata,
but it is based on the actual metadata of the image data.
"""

import math
from typing import overload

from ngio.ome_zarr_meta import (
    AxesHandler,
)
from ngio.ome_zarr_meta.ngio_specs._dataset import Dataset
from ngio.ome_zarr_meta.ngio_specs._pixel_size import PixelSize
from ngio.utils import NgioValueError


def _are_compatible(shape1: int, shape2: int, scaling: float) -> bool:
    """Check if shape2 is consistent with shape1 given pixel sizes.

    Since we only deal with shape discrepancies due to rounding, we
    shape1, needs to be larger than shape2.
    """
    if shape1 < shape2:
        return _are_compatible(shape2, shape1, 1 / scaling)
    expected_shape2 = shape1 * scaling
    expected_shape2_floor = math.floor(expected_shape2)
    expected_shape2_ceil = math.ceil(expected_shape2)
    return shape2 in {expected_shape2_floor, expected_shape2_ceil}


def require_axes_match(reference: "Dimensions", other: "Dimensions") -> None:
    """Check if two Dimensions objects have the same axes.

    Besides the channel axis (which is a special case), all axes must be
    present in both Dimensions objects.

    Args:
        reference (Dimensions): The reference dimensions object to compare against.
        other (Dimensions): The other dimensions object to compare against.

    Raises:
        NgioValueError: If the axes do not match.
    """
    for s_axis in reference.axes_handler.axes:
        if s_axis.axis_type == "channel":
            continue
        o_axis = other.axes_handler.get_axis(s_axis.name)
        if o_axis is None:
            raise NgioValueError(
                f"Axes do not match. The axis {s_axis.name} "
                f"is not present in either dimensions."
            )
    # Check for axes present in the other dimensions but not in this one
    for o_axis in other.axes_handler.axes:
        if o_axis.axis_type == "channel":
            continue
        s_axis = reference.axes_handler.get_axis(o_axis.name)
        if s_axis is None:
            raise NgioValueError(
                f"Axes do not match. The axis {o_axis.name} "
                f"is not present in either dimensions."
            )


def check_if_axes_match(reference: "Dimensions", other: "Dimensions") -> bool:
    """Check if two Dimensions objects have the same axes.

    Besides the channel axis (which is a special case), all axes must be
    present in both Dimensions objects.

    Args:
        reference (Dimensions): The reference dimensions object to compare against.
        other (Dimensions): The other dimensions object to compare against.

    Returns:
        bool: True if the axes match, False otherwise.
    """
    try:
        require_axes_match(reference, other)
        return True
    except NgioValueError:
        return False


def require_dimensions_match(
    reference: "Dimensions", other: "Dimensions", allow_singleton: bool = False
) -> None:
    """Check if two Dimensions objects have the same axes and dimensions.

    Besides the channel axis, all axes must have the same dimension in
    both images.

    Args:
        reference (Dimensions): The reference dimensions object to compare against.
        other (Dimensions): The other dimensions object to compare against.
        allow_singleton (bool): Whether to allow singleton dimensions to be
            different. For example, if the input image has shape
            (5, 100, 100) and the label has shape (1, 100, 100).

    Raises:
        NgioValueError: If the dimensions do not match.
    """
    require_axes_match(reference, other)
    for r_axis in reference.axes_handler.axes:
        if r_axis.axis_type == "channel":
            continue
        o_axis = other.axes_handler.get_axis(r_axis.name)
        assert o_axis is not None  # already checked in assert_axes_match

        r_dim = reference.get(r_axis.name, default=1)
        o_dim = other.get(o_axis.name, default=1)

        if r_dim != o_dim:
            if allow_singleton and (r_dim == 1 or o_dim == 1):
                continue
            raise NgioValueError(
                f"Dimensions do not match for axis "
                f"{r_axis.name}. Got {r_dim} and {o_dim}."
            )


def check_if_dimensions_match(
    reference: "Dimensions", other: "Dimensions", allow_singleton: bool = False
) -> bool:
    """Check if two Dimensions objects have the same axes and dimensions.

    Besides the channel axis, all axes must have the same dimension in
    both images.

    Args:
        reference (Dimensions): The reference dimensions object to compare against.
        other (Dimensions): The other dimensions object to compare against.
        allow_singleton (bool): Whether to allow singleton dimensions to be
            different. For example, if the input image has shape
            (5, 100, 100) and the label has shape (1, 100, 100).

    Returns:
        bool: True if the dimensions match, False otherwise.
    """
    try:
        require_dimensions_match(reference, other, allow_singleton)
        return True
    except NgioValueError:
        return False


def require_rescalable(reference: "Dimensions", other: "Dimensions") -> None:
    """Assert that two images can be rescaled.

    For this to be true, the images must have the same axes, and
    the pixel sizes must be compatible (i.e. one can be scaled to the other).

    Args:
        reference (Dimensions): The reference dimensions object to compare against.
        other (Dimensions): The other dimensions object to compare against.

    """
    require_axes_match(reference, other)
    for ax_r in reference.axes_handler.axes:
        if ax_r.axis_type == "channel":
            continue
        ax_o = other.axes_handler.get_axis(ax_r.name)
        assert ax_o is not None, "Axes do not match."
        px_r = reference.pixel_size.get(ax_r.name, default=1.0)
        px_o = other.pixel_size.get(ax_o.name, default=1.0)
        shape_r = reference.get(ax_r.name, default=1)
        shape_o = other.get(ax_o.name, default=1)
        scale = px_r / px_o
        if not _are_compatible(
            shape1=shape_r,
            shape2=shape_o,
            scaling=scale,
        ):
            raise NgioValueError(
                f"Reference image with shape {reference.shape}, "
                f"and pixel size {reference.pixel_size}, "
                f"cannot be rescaled to "
                f"image with shape {other.shape} "
                f"and pixel size {other.pixel_size}. "
            )


def check_if_rescalable(reference: "Dimensions", other: "Dimensions") -> bool:
    """Check if two images can be rescaled.

    For this to be true, the images must have the same axes, and
    the pixel sizes must be compatible (i.e. one can be scaled to the other).

    Args:
        reference (Dimensions): The reference dimensions object to compare against.
        other (Dimensions): The other dimensions object to compare against.

    Returns:
        bool: True if the images can be rescaled, False otherwise.
    """
    try:
        require_rescalable(reference, other)
        return True
    except NgioValueError:
        return False


class Dimensions:
    """Dimension metadata Handling Class.

    This class is used to handle and manipulate dimension metadata.
    It provides methods to access and validate dimension information,
    such as shape, axes, and properties like is_2d, is_3d, is_time_series, etc.
    """

    require_axes_match = require_axes_match
    check_if_axes_match = check_if_axes_match
    require_dimensions_match = require_dimensions_match
    check_if_dimensions_match = check_if_dimensions_match
    require_rescalable = require_rescalable
    check_if_rescalable = check_if_rescalable

    def __init__(
        self,
        shape: tuple[int, ...],
        chunks: tuple[int, ...],
        dataset: Dataset,
    ) -> None:
        """Create a Dimension object from a Zarr array.

        Args:
            shape: The shape of the Zarr array.
            chunks: The chunks of the Zarr array.
            dataset: The dataset object.
        """
        self._shape = shape
        self._chunks = chunks
        self._axes_handler = dataset.axes_handler
        self._pixel_size = dataset.pixel_size

        if len(self._shape) != len(self._axes_handler.axes):
            raise NgioValueError(
                "The number of dimensions must match the number of axes. "
                f"Expected Axis {self._axes_handler.axes_names} but got shape "
                f"{self._shape}."
            )

    def __str__(self) -> str:
        """Return the string representation of the object."""
        dims = ", ".join(
            f"{ax.name}: {s}"
            for ax, s in zip(self._axes_handler.axes, self._shape, strict=True)
        )
        return f"Dimensions({dims})"

    def __repr__(self) -> str:
        """Return the string representation of the object."""
        return str(self)

    @property
    def axes_handler(self) -> AxesHandler:
        """Return the axes handler object."""
        return self._axes_handler

    @property
    def pixel_size(self) -> PixelSize:
        """Return the pixel size object."""
        return self._pixel_size

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape as a tuple."""
        return self._shape

    @property
    def chunks(self) -> tuple[int, ...]:
        """Return the chunks as a tuple."""
        return self._chunks

    @property
    def axes(self) -> tuple[str, ...]:
        """Return the axes as a tuple of strings."""
        return self.axes_handler.axes_names

    @property
    def is_time_series(self) -> bool:
        """Return whether the image is a time series."""
        if self.get("t", default=1) == 1:
            return False
        return True

    @property
    def is_2d(self) -> bool:
        """Return whether the image is 2D."""
        if self.get("z", default=1) != 1:
            return False
        return True

    @property
    def is_2d_time_series(self) -> bool:
        """Return whether the image is a 2D time series."""
        return self.is_2d and self.is_time_series

    @property
    def is_3d(self) -> bool:
        """Return whether the image is 3D."""
        return not self.is_2d

    @property
    def is_3d_time_series(self) -> bool:
        """Return whether the image is a 3D time series."""
        return self.is_3d and self.is_time_series

    @property
    def is_multi_channels(self) -> bool:
        """Return whether the image has multiple channels."""
        if self.get("c", default=1) == 1:
            return False
        return True

    @overload
    def get(self, axis_name: str, default: None = None) -> int | None:
        pass

    @overload
    def get(self, axis_name: str, default: int) -> int:
        pass

    def get(self, axis_name: str, default: int | None = None) -> int | None:
        """Return the dimension/shape of the given axis name.

        Args:
            axis_name: The name of the axis (either canonical or non-canonical).
            default: The default value to return if the axis does not exist.
        """
        index = self.axes_handler.get_index(axis_name)
        if index is None:
            return default
        return self._shape[index]



import pytest

from ngio.common import Dimensions
from ngio.ome_zarr_meta import AxesHandler, Dataset
from ngio.ome_zarr_meta.ngio_specs import Axis
from ngio.utils import NgioValueError


@pytest.mark.parametrize(
    "axes_names",
    [
        ["x", "y", "z", "c"],
        ["x", "y", "c"],
        ["z", "c", "x", "y"],
        ["t", "z", "c", "x", "y"],
        ["x", "y", "z", "t"],
    ],
)
def test_dimensions(axes_names):
    axes = [Axis(name=name) for name in axes_names]
    canonic_dim_dict = dict(zip("tczyx", (2, 3, 4, 5, 6), strict=True))
    dim_dict = {ax: canonic_dim_dict.get(ax, 1) for ax in axes_names}

    shape = tuple(dim_dict.get(ax) for ax in axes_names)
    shape = tuple(s for s in shape if s is not None)

    ax_mapper = AxesHandler(axes=axes)

    ds = Dataset(
        path="0",
        axes_handler=ax_mapper,
        scale=[1.0] * len(axes_names),
        translation=[0.0] * len(axes_names),
    )

    dims = Dimensions(shape=shape, chunks=shape, dataset=ds)

    assert isinstance(dims.__repr__(), str)

    for ax, s in dim_dict.items():
        assert dims.get(ax) == s

    assert dims.axes_handler.get_index("x") == axes_names.index("x")
    assert dims.axes_handler.get_index("-") is None

    if dim_dict.get("z", 1) > 1:
        assert dims.is_3d

    if dim_dict.get("c", 1) > 1:
        assert dims.is_multi_channels

    if dim_dict.get("t", 1) > 1:
        assert dims.is_time_series

    if dim_dict.get("z", 1) > 1 and dim_dict.get("t", 1) > 1:
        assert dims.is_3d_time_series

    if dim_dict.get("z", 1) == 1 and dim_dict.get("t", 1) == 1:
        assert dims.is_2d

    if dim_dict.get("z", 1) == 1 and dim_dict.get("t", 1) > 1:
        assert dims.is_2d_time_series

    assert dims.shape == shape


def test_dimensions_error():
    axes = [Axis(name="x"), Axis(name="y")]
    shape = (1, 2, 3)
    ax_handler = AxesHandler(axes=axes)
    ds = Dataset(
        path="0",
        axes_handler=ax_handler,
        scale=[1.0] * len(axes),
        translation=[0.0] * len(axes),
    )

    with pytest.raises(NgioValueError):
        Dimensions(shape=shape, chunks=shape, dataset=ds)

    shape = (3, 4)
    dims = Dimensions(shape=shape, chunks=shape, dataset=ds)

    assert dims.get("c", default=None) is None
    assert not dims.is_3d
    assert not dims.is_multi_channels
    assert not dims.is_time_series
    assert not dims.is_3d_time_series
    assert not dims.is_2d_time_series


def test_dimensions_checks():
    axes = [Axis(name="c"), Axis(name="z"), Axis(name="x"), Axis(name="y")]
    shape = (3, 4, 8, 8)
    ax_handler = AxesHandler(axes=axes)
    ds = Dataset(
        path="0",
        axes_handler=ax_handler,
        scale=[1.0] * len(axes),
        translation=[0.0] * len(axes),
    )
    dims_ref = Dimensions(shape=shape, chunks=shape, dataset=ds)

    # Test with rescalable dimensions
    # Axes matches, dimensions do not match, but are rescalable
    axes = [Axis(name="c"), Axis(name="z"), Axis(name="x"), Axis(name="y")]
    shape = (3, 4, 16, 16)
    ax_handler = AxesHandler(axes=axes)
    ds = Dataset(
        path="0",
        axes_handler=ax_handler,
        scale=[1.0, 1.0, 0.5, 0.5],
        translation=[0.0] * len(axes),
    )
    dims_other = Dimensions(shape=shape, chunks=shape, dataset=ds)

    dims_ref.require_axes_match(dims_other)
    assert dims_ref.check_if_axes_match(dims_other)

    with pytest.raises(NgioValueError):
        dims_ref.require_dimensions_match(dims_other, allow_singleton=False)
    assert not dims_ref.check_if_dimensions_match(dims_other, allow_singleton=False)

    dims_ref.require_rescalable(dims_other)
    assert dims_ref.check_if_rescalable(dims_other)

    # Test with non-matching axes
    # Axes do not match

    axes = [Axis(name="x"), Axis(name="y")]
    shape = (8, 8)
    ax_handler = AxesHandler(axes=axes)
    ds = Dataset(
        path="0",
        axes_handler=ax_handler,
        scale=[1.0] * len(axes),
        translation=[0.0] * len(axes),
    )
    dims_other = Dimensions(shape=shape, chunks=shape, dataset=ds)

    with pytest.raises(NgioValueError):
        dims_ref.require_axes_match(dims_other)
    assert not dims_ref.check_if_axes_match(dims_other)

    # Axes match, dimensions do (with singleton allowed)
    axes = [Axis(name="z"), Axis(name="x"), Axis(name="y")]
    shape = (1, 8, 8)
    ax_handler = AxesHandler(axes=axes)
    ds = Dataset(
        path="0",
        axes_handler=ax_handler,
        scale=[1.0] * len(axes),
        translation=[0.0] * len(axes),
    )
    dims_other = Dimensions(shape=shape, chunks=shape, dataset=ds)
    dims_ref.require_axes_match(dims_other)
    assert dims_ref.check_if_axes_match(dims_other)

    assert not dims_ref.check_if_dimensions_match(dims_other, allow_singleton=False)
    dims_ref.require_dimensions_match(dims_other, allow_singleton=True)
    assert dims_ref.check_if_dimensions_match(dims_other, allow_singleton=True)



"""Abstract class for handling OME-NGFF images."""

from collections.abc import Mapping, Sequence
from typing import Any, Literal

import numpy as np
import PIL.Image
from zarr.core.array import CompressorLike

from ngio.common._pyramid import ChunksLike, ShardsLike
from ngio.common._synt_images_utils import fit_to_shape
from ngio.images._ome_zarr_container import OmeZarrContainer, create_ome_zarr_from_array
from ngio.ome_zarr_meta.ngio_specs import (
    Channel,
    DefaultNgffVersion,
    NgffVersions,
)
from ngio.resources import AVAILABLE_SAMPLES, SampleInfo, get_sample_info
from ngio.tables import (
    DefaultTableBackend,
    TableBackend,
)
from ngio.utils import (
    StoreOrGroup,
)


def create_synthetic_ome_zarr(
    store: StoreOrGroup,
    shape: Sequence[int],
    reference_sample: AVAILABLE_SAMPLES | SampleInfo = "Cardiomyocyte",
    levels: int | list[str] = 5,
    table_backend: TableBackend = DefaultTableBackend,
    scaling_factors: Sequence[float] | Literal["auto"] = "auto",
    axes_names: Sequence[str] | None = None,
    channels_meta: Sequence[str | Channel] | None = None,
    ngff_version: NgffVersions = DefaultNgffVersion,
    chunks: ChunksLike = "auto",
    shards: ShardsLike | None = None,
    dimension_separator: Literal[".", "/"] = "/",
    compressors: CompressorLike = "auto",
    extra_array_kwargs: Mapping[str, Any] | None = None,
    overwrite: bool = False,
) -> OmeZarrContainer:
    """Create a synthetic OME-Zarr image with the given shape and metadata.

    Args:
        store (StoreOrGroup): The Zarr store or group to create the image in.
        shape (Sequence[int]): The shape of the image.
        reference_sample (AVAILABLE_SAMPLES | SampleInfo): The reference sample to use.
            Defaults to "Cardiomyocyte".
        levels (int | list[str]): The number of levels in the pyramid or a list of
            level names. Defaults to 5.
        table_backend (TableBackend): Table backend to be used to store tables.
            Defaults to DefaultTableBackend.
        scaling_factors (Sequence[float] | Literal["auto"]): The down-scaling factors
            for the pyramid levels. Defaults to "auto".
        axes_names (Sequence[str] | None): The names of the axes. If None the
            canonical names are used. Defaults to None.
        channels_meta (Sequence[str | Channel] | None): The channels metadata.
            Defaults to None.
        ngff_version (NgffVersions): The version of the OME-Zarr specification.
            Defaults to DefaultNgffVersion.
        chunks (ChunksLike): The chunk shape. Defaults to "auto".
        shards (ShardsLike | None): The shard shape. Defaults to None.
        dimension_separator (Literal[".", "/"]): The separator to use for
            dimensions. Defaults to "/".
        compressors (CompressorLike): The compressors to use. Defaults to "auto".
        extra_array_kwargs (Mapping[str, Any] | None): Extra arguments to pass to
            the zarr array creation. Defaults to None.
        overwrite (bool): Whether to overwrite an existing image. Defaults to False.
    """
    if isinstance(reference_sample, str):
        sample_info = get_sample_info(reference_sample)
    else:
        sample_info = reference_sample

    raw = np.asarray(PIL.Image.open(sample_info.img_path))
    raw = fit_to_shape(arr=raw, out_shape=tuple(shape))
    raw = raw / np.max(raw) * (2**16 - 1)
    raw = raw.astype(np.uint16)
    ome_zarr = create_ome_zarr_from_array(
        store=store,
        array=raw,
        pixelsize=sample_info.xy_pixelsize,
        z_spacing=sample_info.z_spacing,
        time_spacing=sample_info.time_spacing,
        levels=levels,
        space_unit=sample_info.space_unit,
        time_unit=sample_info.time_unit,
        axes_names=axes_names,
        channels_meta=channels_meta,
        scaling_factors=scaling_factors,
        extra_array_kwargs=extra_array_kwargs,
        name=sample_info.name,
        chunks=chunks,
        shards=shards,
        overwrite=overwrite,
        dimension_separator=dimension_separator,
        compressors=compressors,
        ngff_version=ngff_version,
    )

    image = ome_zarr.get_image()
    well_table = image.build_image_roi_table()
    ome_zarr.add_table("well_ROI_table", table=well_table, backend=table_backend)

    for label_info in sample_info.labels:
        ome_zarr.derive_label(name=label_info.name)
        label = ome_zarr.get_label(name=label_info.name)

        ref_label = np.asarray(PIL.Image.open(label_info.label_path))
        ref_label = ref_label.astype(label_info.dtype)

        ref_label = fit_to_shape(
            arr=ref_label,
            out_shape=label.shape,
            ensure_unique_info=label_info.ensure_unique_labels,
        )
        ref_label = ref_label.astype(np.uint32)
        label.set_array(ref_label)
        label.consolidate()

        if label_info.create_masking_table:
            masking_table = label.build_masking_roi_table()
            ome_zarr.add_table(
                name=f"{label_info.name}_masking_table",
                table=masking_table,
                backend=table_backend,
            )

    return ome_zarr



"""ngio internal specs module.

Since the OME-Zarr specification are still evolving, this module provides a
set of classes to internally handle the metadata.

This models can be tr
"""

from ngio.ome_zarr_meta.ngio_specs._axes import (
    AxesHandler,
    AxesSetup,
    Axis,
    AxisType,
    DefaultSpaceUnit,
    DefaultTimeUnit,
    SpaceUnits,
    TimeUnits,
    build_canonical_axes_handler,
    canonical_axes_order,
    canonical_label_axes_order,
)
from ngio.ome_zarr_meta.ngio_specs._channels import (
    Channel,
    ChannelsMeta,
    ChannelVisualisation,
    NgioColors,
    default_channel_name,
)
from ngio.ome_zarr_meta.ngio_specs._dataset import Dataset
from ngio.ome_zarr_meta.ngio_specs._ngio_hcs import (
    ImageInWellPath,
    NgioPlateMeta,
    NgioWellMeta,
    path_in_well_validation,
)
from ngio.ome_zarr_meta.ngio_specs._ngio_image import (
    DefaultNgffVersion,
    ImageLabelSource,
    NgffVersions,
    NgioImageLabelMeta,
    NgioImageMeta,
    NgioLabelMeta,
    NgioLabelsGroupMeta,
)
from ngio.ome_zarr_meta.ngio_specs._pixel_size import PixelSize

__all__ = [
    "AxesHandler",
    "AxesSetup",
    "Axis",
    "AxisType",
    "Channel",
    "ChannelVisualisation",
    "ChannelsMeta",
    "Dataset",
    "DefaultNgffVersion",
    "DefaultSpaceUnit",
    "DefaultTimeUnit",
    "ImageInWellPath",
    "ImageLabelSource",
    "NgffVersions",
    "NgioColors",
    "NgioImageLabelMeta",
    "NgioImageMeta",
    "NgioLabelMeta",
    "NgioLabelsGroupMeta",
    "NgioPlateMeta",
    "NgioWellMeta",
    "PixelSize",
    "SpaceUnits",
    "TimeUnits",
    "build_canonical_axes_handler",
    "canonical_axes_order",
    "canonical_label_axes_order",
    "default_channel_name",
    "path_in_well_validation",
]



"""Protocol for table backends handlers."""

from typing import Literal, Protocol

from anndata import AnnData
from pandas import DataFrame
from polars import DataFrame as PolarsDataFrame
from polars import LazyFrame

from ngio.tables.backends._anndata import AnnDataBackend, AnnDataBackendV1
from ngio.tables.backends._csv import CsvTableBackend
from ngio.tables.backends._json import JsonTableBackend
from ngio.tables.backends._parquet import ParquetTableBackend
from ngio.tables.backends._utils import TabularData
from ngio.utils import NgioValueError, ZarrGroupHandler


class TableBackendProtocol(Protocol):
    def set_group_handler(
        self,
        group_handler: ZarrGroupHandler,
        index_key: str | None = None,
        index_type: Literal["int", "str"] | None = None,
    ) -> None:
        """Attach a group handler to the backend.

        Index keys and index types are used to ensure that the
        serialization and deserialization of the table
        is consistent across different backends.

        Making sure that this is consistent is
        a duty of the backend implementations.
        """
        ...

    @staticmethod
    def backend_name() -> str:
        """Return the name of the backend.

        As a convention we set name as:
            {backend_name}_v{version}

        Where the version is a integer.
        """
        ...

    @property
    def group_handler(self) -> ZarrGroupHandler:
        """Return the group handler."""
        ...

    @staticmethod
    def implements_anndata() -> bool:
        """Check if the backend implements the anndata protocol.

        If this is True, the backend should implement the
        `write_from_anndata` method.

        AnnData objects are more complex than DataFrames,
        so if this is true the backend should implement the
        full serialization of the AnnData object.

        If this is False, these methods should raise a
        `NotImplementedError`.
        """
        ...

    @staticmethod
    def implements_pandas() -> bool:
        """Check if the backend implements the pandas protocol.

        If this is True, the backend should implement the
        `write_from_dataframe` methods.

        If this is False, these methods should raise a
        `NotImplementedError`.
        """
        ...

    @staticmethod
    def implements_polars() -> bool:
        """Check if the backend implements the polars protocol.

        If this is True, the backend should implement the
        `write_from_polars` methods.

        If this is False, these methods should raise a
        `NotImplementedError`.
        """
        ...

    def load_as_anndata(self) -> AnnData:
        """Load the table as an AnnData object."""
        ...

    def load_as_pandas_df(self) -> DataFrame:
        """Load the table as a pandas DataFrame."""
        ...

    def load_as_polars_lf(self) -> LazyFrame:
        """Load the table as a polars LazyFrame."""
        ...

    def load(self) -> TabularData:
        """The default load method.

        This method will be default way to load the table
        from the backend. This method should wrap one of the
        `load_as_anndata`, `load_as_dataframe` or `load_as_polars`
        methods depending on the backend implementation.
        """
        ...

    def write_from_pandas(self, table: DataFrame) -> None:
        """Serialize the table from a pandas DataFrame."""
        ...

    def write_from_anndata(self, table: AnnData) -> None:
        """Serialize the table from an AnnData object."""
        ...

    def write_from_polars(self, table: LazyFrame | PolarsDataFrame) -> None:
        """Serialize the table from a polars DataFrame or LazyFrame."""
        ...

    def write(
        self,
        table_data: DataFrame | AnnData | PolarsDataFrame | LazyFrame,
        metadata: dict[str, str] | None = None,
    ) -> None:
        """This is a generic write method.

        Will call the appropriate write method
        depending on the type of the table.

        Moreover it will also write the metadata
        if provided, and the backend methadata

        the backend should write in the zarr group attributes
            - backend: the backend name (self.backend_name())
            - index_key: the index key
            - index_type: the index type

        """


class ImplementedTableBackends:
    """A class to manage the available table backends."""

    _instance = None
    _implemented_backends: dict[str, type[TableBackendProtocol]]

    def __new__(cls):
        """Create a new instance of the class if it does not exist."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance._implemented_backends = {}
        return cls._instance

    @property
    def available_backends(self) -> list[str]:
        """Return the available table backends."""
        return list(self._implemented_backends.keys())

    def get_backend(
        self,
        *,
        group_handler: ZarrGroupHandler,
        backend_name: str = "anndata",
        index_key: str | None = None,
        index_type: Literal["int", "str"] | None = None,
    ) -> TableBackendProtocol:
        """Try to get a handler for the given store based on the metadata version."""
        if backend_name not in self._implemented_backends:
            raise NgioValueError(f"Table backend {backend_name} not implemented.")
        backend = self._implemented_backends[backend_name]()
        backend.set_group_handler(
            group_handler=group_handler, index_key=index_key, index_type=index_type
        )
        return backend

    def _add_backend(
        self,
        table_backend: type[TableBackendProtocol],
        name: str,
        overwrite: bool = False,
    ) -> None:
        """Register a new handler."""
        if name in self._implemented_backends and not overwrite:
            raise NgioValueError(
                f"Table backend {name} already implemented. "
                "Use the `overwrite=True` parameter to overwrite it."
            )
        self._implemented_backends[name] = table_backend

    def add_backend(
        self,
        table_backend: type[TableBackendProtocol],
        overwrite: bool = False,
        aliases: list[str] | None = None,
    ) -> None:
        """Register a new handler."""
        self._add_backend(
            table_backend=table_backend,
            name=table_backend.backend_name(),
            overwrite=overwrite,
        )
        if aliases is not None:
            for alias in aliases:
                self._add_backend(
                    table_backend=table_backend, name=alias, overwrite=overwrite
                )


ImplementedTableBackends().add_backend(AnnDataBackend)
ImplementedTableBackends().add_backend(AnnDataBackendV1)
ImplementedTableBackends().add_backend(
    JsonTableBackend, aliases=["experimental_json_v1"]
)
ImplementedTableBackends().add_backend(CsvTableBackend, aliases=["experimental_csv_v1"])
ImplementedTableBackends().add_backend(
    ParquetTableBackend, aliases=["experimental_parquet_v1"]
)

TableBackend = Literal["anndata", "json", "csv", "parquet"] | str | TableBackendProtocol
DefaultTableBackend = "anndata_v1"



from pathlib import Path

from ngio import NgffVersions, create_empty_ome_zarr

DATA_DIR = Path("tests/data")

IMAGE_SPECS = [
    {
        "name": "test_image_tcyx.zarr",
        "shape": (4, 3, 64, 64),
        "axes": "tcyx",
    },
    {
        "name": "test_image_tczyx.zarr",
        "shape": (4, 2, 10, 64, 64),
        "axes": "tczyx",
    },
    {
        "name": "test_image_zyx.zarr",
        "shape": (10, 64, 64),
        "axes": "zyx",
    },
    {
        "name": "test_image_c1yx.zarr",
        "shape": (2, 1, 64, 64),
        "axes": "czyx",
    },
    {
        "name": "test_image_tyx.zarr",
        "shape": (4, 64, 64),
        "axes": "tyx",
    },
    {
        "name": "test_image_tzyx.zarr",
        "shape": (4, 10, 64, 64),
        "axes": "tzyx",
    },
    {
        "name": "test_image_cyx.zarr",
        "shape": (2, 64, 64),
        "axes": "cyx",
    },
    {
        "name": "test_image_yx.zarr",
        "shape": (64, 64),
        "axes": "yx",
    },
    {
        "name": "test_image_czyx.zarr",
        "shape": (2, 10, 64, 64),
        "axes": "czyx",
    },
]


def create_test_images_dataset(version: NgffVersions) -> None:
    version_str = "".join(version.split("."))
    base_dir = DATA_DIR / f"v{version_str}" / "images"
    base_dir.mkdir(parents=True, exist_ok=True)
    for spec in IMAGE_SPECS:
        image_path = base_dir / spec["name"]
        ome_zarr = create_empty_ome_zarr(
            store=image_path,
            xy_pixelsize=0.5,
            shape=spec["shape"],
            axes_names=spec["axes"],
            ngff_version=version,
            levels=2,
            overwrite=True,
        )
        ome_zarr.derive_label("label")
        well_roi_table = ome_zarr.build_image_roi_table()
        ome_zarr.add_table(name="well_ROI_table", table=well_roi_table, backend="csv")


if __name__ == "__main__":
    for version in ["0.4", "0.5"]:
        create_test_images_dataset(version)  # type: ignore



import numpy as np
import pytest

from ngio.ome_zarr_meta.ngio_specs import (
    AxesHandler,
    AxesSetup,
    Axis,
    AxisType,
    Channel,
    ChannelsMeta,
    ChannelVisualisation,
    Dataset,
    DefaultSpaceUnit,
    DefaultTimeUnit,
    NgioColors,
    NgioImageMeta,
    NgioLabelMeta,
    PixelSize,
)
from ngio.ome_zarr_meta.ngio_specs._channels import valid_hex_color


@pytest.mark.parametrize(
    "axes, axes_setup, allow_non_canonical_axes, strict_canonical_order",
    [
        (
            ["t", "c", "z", "y", "x"],
            None,
            False,
            True,
        ),
        (
            ["c", "t", "z", "y", "x"],
            None,
            False,
            False,
        ),
        (
            ["c", "t", "z", "y", "X"],
            AxesSetup(x="X"),
            False,
            False,
        ),
        (
            ["y", "X"],
            AxesSetup(x="X"),
            False,
            False,
        ),
        (
            ["weird", "y", "X"],
            AxesSetup(x="X", others=["weird"]),
            True,
            False,
        ),
    ],
)
def test_axes_base(axes, axes_setup, allow_non_canonical_axes, strict_canonical_order):
    _axes = [Axis(name=name) for name in axes]
    mapper = AxesHandler(
        axes=_axes,
        axes_setup=axes_setup,
        allow_non_canonical_axes=allow_non_canonical_axes,
        strict_canonical_order=strict_canonical_order,
    )
    for i, ax in enumerate(axes):
        assert mapper.get_index(ax) == i

    assert len(mapper.axes) == len(axes)


@pytest.mark.parametrize(
    "canonical_name, axis_type, unit, expected_type, expected_unit",
    [
        ("x", AxisType.space, None, AxisType.space, "micrometer"),
        ("x", AxisType.time, "second", AxisType.space, "second"),
        ("t", AxisType.time, None, AxisType.time, "second"),
        ("c", AxisType.channel, None, AxisType.channel, None),
    ],
)
def test_axis_cast(canonical_name, axis_type, unit, expected_type, expected_unit):
    ax = Axis(
        name="temp",
        unit=unit,
        axis_type=axis_type,
    )
    ax = ax.canonical_axis_cast(canonical_name)
    assert ax.axis_type == expected_type
    assert ax.unit == expected_unit


def test_axes_fail():
    with pytest.raises(ValueError):
        AxesHandler(
            axes=[Axis(name="x")],
            axes_setup=AxesSetup(x="X"),
            allow_non_canonical_axes=False,
            strict_canonical_order=False,
        )

    with pytest.raises(ValueError):
        AxesHandler(
            axes=[Axis(name="x")],
            axes_setup=AxesSetup(x="x"),
            allow_non_canonical_axes=True,
            strict_canonical_order=True,
        )

    with pytest.raises(ValueError):
        AxesHandler(
            axes=[
                Axis(name="x"),
                Axis(name="x"),
            ],
            axes_setup=None,
            allow_non_canonical_axes=False,
            strict_canonical_order=True,
        )

    with pytest.raises(ValueError):
        AxesHandler(
            axes=[
                Axis(name="x"),
                Axis(name="z"),
            ],
            axes_setup=None,
            allow_non_canonical_axes=False,
            strict_canonical_order=True,
        )

    with pytest.raises(ValueError):
        AxesHandler(
            axes=[
                Axis(name="weird"),
                Axis(name="y"),
                Axis(name="x"),
            ],
            axes_setup=AxesSetup(others=["weird"]),
            allow_non_canonical_axes=False,
            strict_canonical_order=False,
        )


def test_pixel_size():
    ps_dict = {"x": 0.5, "y": 0.5, "z": 1.0, "t": 1.0}
    ps_1 = PixelSize(**ps_dict, space_unit=DefaultSpaceUnit, time_unit=DefaultTimeUnit)
    assert ps_1.as_dict() == ps_dict
    assert ps_1.zyx == (1.0, 0.5, 0.5)
    assert ps_1.yx == (0.5, 0.5)
    assert ps_1.voxel_volume == 0.25
    assert ps_1.xy_plane_area == 0.25
    assert ps_1.time_spacing == 1.0

    ps_2 = PixelSize(x=0.5, y=0.5, z=1.0, t=1.0)
    np.testing.assert_allclose(ps_1.distance(ps_2), 0.0)
    ps_3 = PixelSize(x=1.0, y=1.0, z=1.0, t=1.0)
    np.testing.assert_allclose(ps_1.distance(ps_3), np.sqrt(2.0) / 2)

    # Test comparison
    p1 = PixelSize(x=1, y=1, z=0.1243532)
    p2 = PixelSize(x=1, y=1, z=0.1243532)
    assert p1 == p2

    p_small = PixelSize(x=0.1, y=0.1, z=0.1)
    p_large = PixelSize(x=2, y=2, z=2)
    assert p_small < p_large


def test_dataset():
    axes = [
        Axis(name="t", axis_type=AxisType.time, unit=DefaultTimeUnit),
        Axis(name="c", axis_type=AxisType.channel),
        Axis(name="z"),
        Axis(name="y"),
        Axis(name="x"),
    ]

    axes_handler = AxesHandler(
        axes=axes,
        axes_setup=AxesSetup(),
        allow_non_canonical_axes=False,
        strict_canonical_order=True,
    )

    scale = [1.0, 1.0, 1.0, 0.5, 0.5]
    translation = [0.0, 0.0, 0.0, 0.0, 0.0]
    ds = Dataset(
        path="0",
        axes_handler=axes_handler,
        scale=scale,
        translation=translation,
    )

    assert ds.path == "0"
    assert ds.axes_handler.get_index("x") == 4
    assert ds.scale == tuple(scale)
    assert ds.translation == tuple(translation)

    ps = ds.pixel_size
    assert ps.x == 0.5
    assert ps.y == 0.5
    assert ps.z == 1.0
    assert ps.t == 1.0


def test_dataset_fail():
    axes = [
        Axis(name="y", unit="centimeter"),
        Axis(name="x", unit="micrometer"),
    ]
    axes_handler = AxesHandler(
        axes=axes,
        axes_setup=AxesSetup(),
        allow_non_canonical_axes=False,
        strict_canonical_order=True,
    )
    ds = Dataset(
        path="0",
        axes_handler=axes_handler,
        scale=[0.5, 0.5],
        translation=[0.0, 0.0],
    )

    assert ds.axes_handler.time_unit is None

    with pytest.raises(ValueError):
        assert ds.axes_handler.space_unit == "micrometer"


def test_channels():
    channels = ChannelsMeta.default_init(
        labels=["DAPI", "GFP", "RFP"],
    )
    assert len(channels.channels) == 3
    assert channels.channels[0].label == "DAPI"
    assert channels.channels[0].wavelength_id == "DAPI"
    assert channels.channels[0].channel_visualisation.color == NgioColors.dapi.value

    channels = ChannelsMeta.default_init(labels=4)
    assert len(channels.channels) == 4
    assert channels.channels[0].label == "channel_0"
    assert channels.channels[0].wavelength_id == "channel_0"
    assert channels.channels[0].channel_visualisation.color == "00FFFF"

    channels = ChannelsMeta.default_init(
        labels=["DAPI", "GFP", "RFP"],
        wavelength_id=["A01_C01", "A02_C02", "A03_C03"],
        colors=["00FF00", "FF0000", "00FFFF"],
        active=[True, False, True],
        end=[100, 200, 300],
        start=[0, 100, 200],
        data_type="float",
    )
    assert len(channels.channels) == 3
    assert channels.channels[0].label == "DAPI"
    assert channels.channels[0].wavelength_id == "A01_C01"
    assert channels.channels[0].channel_visualisation.color == "00FF00"

    with pytest.raises(ValueError):
        ChannelsMeta.default_init(labels=[])

    class Mock:
        pass

    with pytest.raises(ValueError):
        ChannelsMeta.default_init(labels=[Mock()])  # type: ignore

    channel = Channel.default_init(label="DAPI", wavelength_id="A01_C01")
    ChannelsMeta(channels=[channel])

    with pytest.raises(ValueError):
        ChannelsMeta.default_init(labels=["DAPI", "DAPI"])

    with pytest.raises(ValueError):
        ChannelsMeta.default_init(labels=[channel, channel])  # type: ignore

    with pytest.raises(ValueError):
        Channel.default_init(label="DAPI", data_type="color")


def test_ngio_colors():
    assert NgioColors.semi_random_pick(channel_name="DAPI") == NgioColors.dapi
    assert NgioColors.semi_random_pick(channel_name="channel_dapi") == NgioColors.dapi
    assert valid_hex_color(NgioColors.semi_random_pick(channel_name=None))

    for channel, expected in zip(
        ["channel_0", "channel_1", "channel_2", "channel_3"],
        [NgioColors.cyan, NgioColors.magenta, NgioColors.yellow, NgioColors.green],
        strict=True,
    ):
        assert NgioColors.semi_random_pick(channel_name=channel) == expected

    for non_hex_color in ["00000000", "not a color"]:
        assert not valid_hex_color(non_hex_color)

    for color in [None, NgioColors.cyan]:
        ChannelVisualisation.default_init(color=color)

    ChannelVisualisation(color=NgioColors.cyan)

    with pytest.raises(ValueError):
        ChannelVisualisation.default_init(color=[])  # type: ignore


def test_image_meta():
    axes = [
        Axis(name="t", axis_type=AxisType.time, unit=DefaultSpaceUnit),
        Axis(name="c", axis_type=AxisType.channel),
        Axis(name="z"),
        Axis(name="y"),
        Axis(name="x"),
    ]

    axes_handler = AxesHandler(
        axes=axes,
        axes_setup=AxesSetup(),
        allow_non_canonical_axes=False,
        strict_canonical_order=True,
    )
    translation = [0.0, 0.0, 0.0, 0.0, 0.0]
    scale = [1.0, 1.0, 1.0, 0.5, 0.5]

    datasets = []
    for path in range(4):
        ds = Dataset(
            path=str(path),
            axes_handler=axes_handler,
            scale=scale,
            translation=translation,
        )
        datasets.append(ds)
        scale = [s * f for s, f in zip(scale, [1, 1, 1, 2, 2], strict=True)]

    image_meta = NgioImageMeta(version="0.4", name="test", datasets=datasets)

    image_meta.init_channels(labels=["DAPI", "GFP", "RFP"])

    assert image_meta.levels == 4
    assert image_meta.name == "test"
    assert image_meta.version == "0.4"
    assert len(image_meta.scaling_factor()) == 5
    np.testing.assert_allclose(image_meta.scaling_factor(), [1, 1, 1, 2, 2])
    assert image_meta.get_dataset(path="0").path == "0"
    assert image_meta.get_dataset(path="1").path == "1"
    assert image_meta.get_dataset().path == "0"
    assert image_meta.get_dataset(pixel_size=datasets[-1].pixel_size).path == "3"
    channels_meta = image_meta.channels_meta
    assert channels_meta is not None
    assert channels_meta.get_channel_idx(channel_label="DAPI") == 0
    assert channels_meta.get_channel_idx(wavelength_id="DAPI") == 0
    assert channels_meta.channel_labels == ["DAPI", "GFP", "RFP"]


def test_label_meta():
    axes = [
        Axis(name="t", axis_type=AxisType.time, unit=DefaultSpaceUnit),
        Axis(name="z"),
        Axis(name="y"),
        Axis(name="x"),
    ]
    axes_handler = AxesHandler(
        axes=axes,
        axes_setup=AxesSetup(),
        allow_non_canonical_axes=False,
        strict_canonical_order=True,
    )
    translation = [0.0, 0.0, 0.0, 0.0]
    scale = [1.0, 1.0, 0.5, 0.5]

    datasets = []
    for path in range(4):
        ds = Dataset(
            path=str(path),
            axes_handler=axes_handler,
            scale=scale,
            translation=translation,
        )
        datasets.append(ds)
        scale = [s * f for s, f in zip(scale, [1, 1, 2, 2], strict=True)]

    label_meta = NgioLabelMeta(
        version="0.4",
        name="test",
        datasets=datasets,
    )
    assert label_meta.source_image == "../../"
    assert label_meta.levels == 4
    assert label_meta.name == "test"
    assert label_meta.version == "0.4"
    np.testing.assert_allclose(label_meta.scaling_factor(), [1, 1, 2, 2])
    assert label_meta.get_dataset(path="0").path == "0"
    assert label_meta.get_dataset(path="1").path == "1"
    assert label_meta.get_dataset().path == "0"
    assert label_meta.get_dataset(pixel_size=datasets[-1].pixel_size).path == "3"


def test_channels_label_meta():
    axes = [
        Axis(name="t", axis_type=AxisType.time, unit=DefaultSpaceUnit),
        Axis(name="c"),
        Axis(name="z"),
        Axis(name="y"),
        Axis(name="x"),
    ]
    axes_handler = AxesHandler(
        axes=axes,
        axes_setup=AxesSetup(),
        allow_non_canonical_axes=False,
        strict_canonical_order=True,
    )
    translation = [0.0, 0.0, 0.0, 0.0, 0.0]
    scale = [1.0, 1.0, 1.0, 0.5, 0.5]

    datasets = []
    for path in range(4):
        ds = Dataset(
            path=str(path),
            axes_handler=axes_handler,
            scale=scale,
            translation=translation,
        )
        datasets.append(ds)
        scale = [s * f for s, f in zip(scale, [1, 1, 1, 2, 2], strict=True)]

    _ = NgioLabelMeta(version="0.4", name="test", datasets=datasets)
